import os
import os.path as osp
import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F

from .. import BaseModel
from .mlp import MLP
from cogdl.utils import spmm


def average_neighbor_features(graph, feats, nhop, norm="sym", style="all"):
    results = []
    if norm == "sym":
        graph.sym_norm()
    elif norm == "row":
        graph.row_norm()
    else:
        raise NotImplementedError

    x = feats
    results.append(x)
    for i in range(nhop):
        x = spmm(graph, x)
        if style == "all":
            results.append(x)
    if style != "all":
        results = x
    return results


def entropy(probs):
    eps = 1e-9
    res = -probs * torch.log(probs + eps) - (1 - probs) * torch.log(1 - probs + eps)
    return res


def prepare_feats(dataset, nhop, norm="sym"):
    print("Preprocessing features...")
    dataset_name = dataset.__class__.__name__
    data = dataset.data
    if not osp.exists("./sagn"):
        os.makedirs("sagn", exist_ok=True)

    feat_emb_path = f"./sagn/feats_{dataset_name}_hop_{nhop}.pt"
    is_inductive = data.is_inductive()
    train_nid = data.train_nid

    if is_inductive:
        if osp.exists(feat_emb_path):
            print("Loading existing features")
            feats = torch.load(feat_emb_path)
        else:
            data.train()
            feats_train = average_neighbor_features(data, data.x, nhop, norm=norm, style="all")
            data.eval()
            feats = average_neighbor_features(data, data.x, nhop, norm=norm, style="all")

            # train_nid, val_nid, test_nid = data.train_nid, data.val_nid, data.test_nid
            for i in range(len(feats)):
                feats[i][train_nid] = feats_train[i][train_nid]
                # feats_train, feats_val, feats_test = feats[train_nid], feats[val_nid], feats[test_nid]
                # feats[i] = torch.cat([feats_train, feats_val, feats_test], dim=0)
            torch.save(feats, feat_emb_path)
    else:
        if osp.exists(feat_emb_path):
            feats = torch.load(feat_emb_path)
        else:
            feats = average_neighbor_features(data, data.x, nhop, norm=norm, style="all")
    print("Preprocessing features done...")
    return feats


def prepare_labels(dataset, stage, nhop, threshold, probs=None, norm="row", load_emb=False):
    dataset_name = dataset.__class__.__name__
    data = dataset.data
    is_inductive = data.is_inductive()
    multi_label = len(data.y.shape) > 1

    device = data.x.device
    num_classes = data.num_classes
    train_nid = data.train_nid
    val_nid = data.val_nid
    test_nid = data.test_nid

    if not osp.exists("./sagn"):
        os.makedirs("sagn", exist_ok=True)
    label_emb_path = f"./sagn/label_emb_{dataset_name}.pt"
    # teacher_prob_path = f"./sagn/teacher_prob_{dataset_name}.pt"
    teacher_probs = probs

    if stage > 0 and probs is not None:
        # teacher_probs = torch.load(teacher_prob_path)
        node_idx = torch.cat([train_nid, val_nid, test_nid], dim=0)
        if multi_label:
            threshold = -threshold * np.log(threshold) - (1 - threshold) * np.log(1 - threshold)
            entropy_distribution = entropy(teacher_probs)
            confident_nid = torch.arange(len(teacher_probs))[(entropy_distribution.mean(1) <= threshold)]
        else:
            confident_nid = torch.arange(len(teacher_probs))[teacher_probs.max(1)[0] > threshold]
        extra_confident_nid = confident_nid[confident_nid >= len(train_nid)]
        confident_nid = node_idx[confident_nid]
        extra_confident_nid = node_idx[extra_confident_nid]

        if multi_label:
            pseudo_labels = teacher_probs
            pseudo_labels[pseudo_labels >= 0.5] = 1
            pseudo_labels[pseudo_labels < 0.5] = 0
            labels_with_pseudos = torch.ones_like(data.y)
        else:
            pseudo_labels = torch.argmax(teacher_probs, dim=1)
            labels_with_pseudos = torch.zeros_like(data.y)
        train_nid_with_pseudos = np.union1d(train_nid.cpu().numpy(), confident_nid.cpu().numpy())
        train_nid_with_pseudos = torch.from_numpy(train_nid_with_pseudos).to(device)
        labels_with_pseudos[train_nid] = data.y[train_nid]
        labels_with_pseudos[extra_confident_nid] = pseudo_labels[extra_confident_nid]
    else:
        # confident_nid = train_nid
        train_nid_with_pseudos = train_nid
        labels_with_pseudos = data.y.clone()
        # teacher_probs = None
        # pseudo_labels = None

    if (not is_inductive) or stage > 0:
        if multi_label:
            label_emb = 0.5 * torch.ones((data.num_nodes, num_classes), device=device)
            label_emb[train_nid_with_pseudos] = labels_with_pseudos.float()[train_nid_with_pseudos]
        else:
            label_emb = torch.zeros((data.num_nodes, num_classes), device=device)
            label_emb[train_nid_with_pseudos] = F.one_hot(
                labels_with_pseudos[train_nid_with_pseudos], num_classes=num_classes
            ).float()
    else:
        label_emb = None

    if is_inductive:
        if osp.exists(label_emb_path) and load_emb:
            label_emb = torch.load(label_emb_path)
        elif label_emb is not None:
            data.train()
            label_emb_train = average_neighbor_features(data, label_emb, nhop, norm=norm, style="last")
            data.eval()
            label_emb = average_neighbor_features(data, label_emb, nhop, norm=norm, style="last")
            label_emb[train_nid] = label_emb_train[train_nid]
            if load_emb:
                torch.save(label_emb, label_emb_path)
    else:
        if osp.exists(label_emb_path) and load_emb:
            label_emb = torch.load(label_emb_path)
        elif label_emb is not None:
            if label_emb is not None:
                label_emb = average_neighbor_features(data, label_emb, nhop, norm=norm, style="last")
            if stage == 0 and load_emb:
                torch.save(label_emb, label_emb_path)

    return label_emb, labels_with_pseudos, train_nid_with_pseudos


class SAGN(BaseModel):
    @staticmethod
    def add_args(parser):
        parser.add_argument("--hidden-size", type=int, default=512)
        parser.add_argument("--negative-slope", type=float, default=0.2)
        parser.add_argument("--dropout", type=float, default=0.5)
        parser.add_argument("--input-drop", type=float, default=0.0)
        parser.add_argument("--attn-drop", type=float, default=0.4)
        parser.add_argument("--nhead", type=int, default=2)
        parser.add_argument("--mlp-layer", type=int, default=4)
        parser.add_argument("--use-labels", action="store_true")
        parser.add_argument("--nhop", type=int, default=4)

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.num_classes,
            args.hidden_size,
            args.nhop,
            args.mlp_layer,
            args.nhead,
            args.dropout,
            args.input_drop,
            args.attn_drop,
            args.negative_slope,
            args.use_labels,
        )

    def __init__(
        self,
        in_feats,
        out_feats,
        hidden_size,
        nhop,
        mlp_layer,
        nhead,
        dropout=0.5,
        input_drop=0.0,
        attn_drop=0.0,
        negative_slope=0.2,
        use_labels=False,
    ):
        super(SAGN, self).__init__()
        self.dropout = dropout
        self.nhead = nhead
        self.hidden_size = hidden_size
        self.attn_dropout = attn_drop
        self.input_dropout = input_drop
        self.use_labels = use_labels
        self.negative_slope = negative_slope

        self.norm = nn.BatchNorm1d(hidden_size)
        self.layers = nn.ModuleList(
            [
                MLP(in_feats, hidden_size * nhead, hidden_size, mlp_layer, norm="batchnorm", dropout=dropout)
                for _ in range(nhop + 1)
            ]
        )

        self.mlp = MLP(hidden_size, out_feats, hidden_size, mlp_layer, norm="batchnorm", dropout=dropout)
        self.res_conn = nn.Linear(in_feats, hidden_size * nhead, bias=False)
        if use_labels:
            self.label_mlp = MLP(out_feats, out_feats, hidden_size, 2 * mlp_layer, norm="batchnorm", dropout=dropout)

        self.attn_l = nn.Parameter(torch.FloatTensor(size=(1, nhead, hidden_size)))
        self.attn_r = nn.Parameter(torch.FloatTensor(size=(1, nhead, hidden_size)))

    def reset_parameters(self):
        gain = nn.init.calculate_gain("relu")
        for layer in self.layers:
            layer.reset_parameters()
        nn.init.xavier_normal_(self.attn_r, gain=gain)
        nn.init.xavier_normal_(self.attn_l, gain=gain)
        if self.use_labels:
            self.label_mlp.reset_parameters()
        self.norm.reset_parameters()
        self.mlp.reset_parameters()

    def forward(self, features, y_emb=None):
        out = 0
        features = [F.dropout(x, p=self.input_dropout, training=self.training) for x in features]
        hidden = [self.layers[i](features[i]).view(-1, self.nhead, self.hidden_size) for i in range(len(features))]
        a_r = (hidden[0] * self.attn_r).sum(dim=-1).unsqueeze(-1)
        a_ls = [(h * self.attn_l).sum(dim=-1).unsqueeze(-1) for h in hidden]
        a = torch.cat([(a_l + a_r).unsqueeze(-1) for a_l in a_ls], dim=-1)
        a = F.leaky_relu(a, negative_slope=self.negative_slope)
        a = F.softmax(a, dim=-1)
        a = F.dropout(a, p=self.attn_dropout, training=self.training)

        for i in range(a.shape[-1]):
            out += hidden[i] * a[:, :, :, i]
        out += self.res_conn(features[0]).view(-1, self.nhead, self.hidden_size)
        out = out.mean(1)
        out = F.relu(self.norm(out))
        out = F.dropout(out, p=self.dropout, training=self.training)
        out = self.mlp(out)

        if self.use_labels and y_emb is not None:
            out += self.label_mlp(y_emb)
        return out



import torch.nn.functional as F
from cogdl.layers import GCNLayer

from .. import BaseModel


class M3S(BaseModel):
    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--hidden-size", type=int, default=64)
        parser.add_argument("--dropout", type=float, default=0)
        parser.add_argument("--num-clusters", type=int, default=50)
        parser.add_argument("--num-stages", type=int, default=10)
        parser.add_argument("--epochs-per-stage", type=int, default=50)
        parser.add_argument("--label-rate", type=float, default=1)
        parser.add_argument("--num-new-labels", type=int, default=2)
        parser.add_argument("--alpha", type=float, default=1)
        parser.add_argument("--approximate", action="store_true")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.hidden_size,
            args.num_classes,
            args.dropout,
        )

    def __init__(self, num_features, hidden_size, num_classes, dropout):
        super(M3S, self).__init__()
        self.dropout = dropout
        self.gcn1 = GCNLayer(num_features, hidden_size)
        self.gcn2 = GCNLayer(hidden_size, num_classes)

    def embed(self, graph):
        graph.sym_norm()
        h = graph.x
        h = self.gcn1(graph, h)
        h = F.relu(F.dropout(h, self.dropout, training=self.training))
        return h.detach().cpu().numpy()

    def forward(self, graph):
        graph.sym_norm()
        h = graph.x
        h = self.gcn1(graph, h)
        h = F.dropout(F.relu(h), self.dropout, training=self.training)
        h = self.gcn2(graph, h)
        return h

    def predict(self, data):
        return self.forward(data)



import numpy as np

from .. import DataWrapper


class NetworkEmbeddingDataWrapper(DataWrapper):
    def __init__(self, dataset):
        super(NetworkEmbeddingDataWrapper, self).__init__()

        self.dataset = dataset
        data = dataset[0]

        num_nodes = data.num_nodes
        num_classes = dataset.num_classes
        if len(data.y.shape) > 1:
            self.label_matrix = data.y
        else:
            self.label_matrix = np.zeros((num_nodes, num_classes), dtype=int)
            self.label_matrix[range(num_nodes), data.y.numpy()] = 1

    def train_wrapper(self):
        return self.dataset.data

    def test_wrapper(self):
        return self.label_matrix



import torch
import torch.nn as nn
import torch.nn.functional as F

from .. import BaseModel
from cogdl.utils import spmm


class Grand(BaseModel):
    """
    Implementation of GRAND in paper `"Graph Random Neural Networks for Semi-Supervised Learning on Graphs"`
    <https://arxiv.org/abs/2005.11079>

    Parameters
    ----------
    nfeat : int
        Size of each input features.
    nhid : int
        Size of hidden features.
    nclass : int
        Number of output classes.
    input_droprate : float
        Dropout rate of input features.
    hidden_droprate : float
        Dropout rate of hidden features.
    use_bn : bool
        Using batch normalization.
    dropnode_rate : float
        Rate of dropping elements of input features
    tem : float
        Temperature to sharpen predictions.
    lam : float
         Proportion of consistency loss of unlabelled data
    order : int
        Order of adjacency matrix
    sample : int
        Number of augmentations for consistency loss
    alpha : float
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--hidden-size", type=int, default=32)
        parser.add_argument("--hidden-dropout", type=float, default=0.5)
        parser.add_argument("--input-dropout", type=float, default=0.5)
        parser.add_argument("--bn", type=bool, default=False)
        parser.add_argument("--dropnode-rate", type=float, default=0.5)
        parser.add_argument("--order", type=int, default=5)
        parser.add_argument("--alpha", type=float, default=0.2)

        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.hidden_size,
            args.num_classes,
            args.input_dropout,
            args.hidden_dropout,
            args.bn,
            args.dropnode_rate,
            args.order,
            args.alpha,
        )

    def __init__(
        self,
        nfeat,
        nhid,
        nclass,
        input_droprate,
        hidden_droprate,
        use_bn,
        dropnode_rate,
        order,
        alpha,
    ):
        super(Grand, self).__init__()
        self.layer1 = nn.Linear(nfeat, nhid)
        self.layer2 = nn.Linear(nhid, nclass)
        self.input_droprate = input_droprate
        self.hidden_droprate = hidden_droprate
        self.bn1 = nn.BatchNorm1d(nfeat)
        self.bn2 = nn.BatchNorm1d(nhid)
        self.use_bn = use_bn
        self.order = order
        self.dropnode_rate = dropnode_rate
        self.alpha = alpha

    def drop_node(self, x):
        n = x.shape[0]
        drop_rates = torch.ones(n) * self.dropnode_rate
        if self.training:
            masks = torch.bernoulli(1.0 - drop_rates).unsqueeze(1)
            x = masks.to(x.device) * x

        else:
            x = x * (1.0 - self.dropnode_rate)
        return x

    def rand_prop(self, graph, x):
        x = self.drop_node(x)

        y = x
        for i in range(self.order):
            x = spmm(graph, x).detach_()
            y.add_(x)
        return y.div_(self.order + 1.0).detach_()

    def normalize_x(self, x):
        row_sum = x.sum(1)
        row_inv = row_sum.pow_(-1)
        row_inv.masked_fill_(row_inv == float("inf"), 0)
        x = x * row_inv[:, None]
        return x

    def forward(self, graph):
        graph.sym_norm()
        x = graph.x
        x = self.normalize_x(x)
        x = self.rand_prop(graph, x)
        if self.use_bn:
            x = self.bn1(x)
        x = F.dropout(x, self.input_droprate, training=self.training)
        x = F.relu(self.layer1(x))
        if self.use_bn:
            x = self.bn2(x)
        x = F.dropout(x, self.hidden_droprate, training=self.training)
        x = self.layer2(x)
        return x

    def predict(self, data):
        return self.forward(data)



import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from actnn.layers import QLinear, QReLU, QBatchNorm1d, QDropout

from cogdl.utils import check_mh_spmm, mh_spmm, edge_softmax, spmm, get_activation, get_norm_layer


class ActGATLayer(nn.Module):
    """
    Sparse version GAT layer, similar to https://arxiv.org/abs/1710.10903
    """

    def __init__(
        self, in_features, out_features, nhead=1, alpha=0.2, attn_drop=0.5, activation=None, residual=False, norm=None
    ):
        super(ActGATLayer, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.alpha = alpha
        self.nhead = nhead

        # self.W = nn.Parameter(torch.FloatTensor(in_features, out_features * nhead))
        self.W = QLinear(in_features, out_features * nhead)

        self.a_l = nn.Parameter(torch.zeros(size=(1, nhead, out_features)))
        self.a_r = nn.Parameter(torch.zeros(size=(1, nhead, out_features)))

        self.dropout = QDropout(attn_drop)
        self.leakyrelu = nn.LeakyReLU(self.alpha)
        self.act = None if activation is None else QReLU()
        self.norm = None if norm is None else QBatchNorm1d(out_features * nhead)

        if residual:
            self.residual = nn.Linear(in_features, out_features * nhead)
        else:
            self.register_buffer("residual", None)
        self.reset_parameters()

    def reset_parameters(self):
        def reset(tensor):
            stdv = math.sqrt(6.0 / (tensor.size(-2) + tensor.size(-1)))
            tensor.data.uniform_(-stdv, stdv)

        reset(self.a_l)
        reset(self.a_r)

        stdv = 1.0 / math.sqrt(self.out_features * self.nhead)
        torch.nn.init.uniform_(self.W.weight, -stdv, stdv)

    def forward(self, graph, x):
        # h = torch.matmul(x, self.W).view(-1, self.nhead, self.out_features)
        h = self.W(x).view(-1, self.nhead, self.out_features)
        h[torch.isnan(h)] = 0.0

        row, col = graph.edge_index
        # Self-attention on the nodes - Shared attention mechanism
        h_l = (self.a_l * h).sum(dim=-1)[row]
        h_r = (self.a_r * h).sum(dim=-1)[col]
        edge_attention = self.leakyrelu(h_l + h_r)
        # edge_attention: E * H
        edge_attention = edge_softmax(graph, edge_attention)
        edge_attention = self.dropout(edge_attention)

        if check_mh_spmm() and next(self.parameters()).device.type != "cpu":
            if self.nhead > 1:
                h_prime = mh_spmm(graph, edge_attention, h)
                out = h_prime.view(h_prime.shape[0], -1)
            else:
                edge_weight = edge_attention.view(-1)
                with graph.local_graph():
                    graph.edge_weight = edge_weight
                    out = spmm(graph, h.squeeze(1), actnn=True)
        else:
            with graph.local_graph():
                h_prime = []
                h = h.permute(1, 0, 2).contiguous()
                for i in range(self.nhead):
                    edge_weight = edge_attention[:, i]
                    graph.edge_weight = edge_weight
                    hidden = h[i]
                    assert not torch.isnan(hidden).any()
                    h_prime.append(spmm(graph, hidden))
            out = torch.cat(h_prime, dim=1)

        if self.residual:
            res = self.residual(x)
            out += res
        if self.norm is not None:
            out = self.norm(out)
        if self.act is not None:
            out = self.act(out)
        return out

    def __repr__(self):
        return self.__class__.__name__ + " (" + str(self.in_features) + " -> " + str(self.out_features) + ")"



from cogdl.models.nn.correct_smooth import CorrectSmooth
from .dgi_mw import DGIModelWrapper
from .gcnmix_mw import GCNMixModelWrapper
from .grace_mw import GRACEModelWrapper
from .grand_mw import GrandModelWrapper
from .graphsage_mw import GraphSAGEModelWrapper
from .unsup_graphsage_mw import UnsupGraphSAGEModelWrapper
from .m3s_mw import M3SModelWrapper
from .mvgrl_mw import MVGRLModelWrapper
from .network_embedding_mw import NetworkEmbeddingModelWrapper
from .node_classification_mw import NodeClfModelWrapper
from .pprgo_mw import PPRGoModelWrapper
from .sagn_mw import SAGNModelWrapper
from .self_auxiliary_mw import SelfAuxiliaryModelWrapper
from .correct_smooth_mw import CorrectSmoothModelWrapper



import torch
import numpy as np
from sklearn.cluster import SpectralClustering

from cogdl.utils import spmm
from .. import BaseModel


class AGC(BaseModel):
    r"""The AGC model from the `"Attributed Graph Clustering via Adaptive Graph Convolution"
    <https://arxiv.org/abs/1906.01210>`_ paper

    Args:
        num_clusters (int) : Number of clusters.
        max_iter     (int) : Max iteration to increase k
    """

    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--num-clusters", type=int, default=7)
        parser.add_argument("--max-iter", type=int, default=10)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(args.num_clusters, args.max_iter, args.cpu)

    def __init__(self, num_clusters, max_iter, cpu):
        super(AGC, self).__init__()

        self.num_clusters = num_clusters
        self.max_iter = max_iter
        self.cpu = cpu

    def forward(self, data):
        device = "cuda" if torch.cuda.is_available() and not self.cpu else "cpu"
        data = data.to(device)
        self.num_nodes = data.x.shape[0]
        graph = data
        graph.add_remaining_self_loops()

        graph.sym_norm()
        graph.edge_weight = data.edge_weight * 0.5

        pre_intra = 1e27
        pre_feat = None
        for t in range(1, self.max_iter + 1):
            x = data.x
            for i in range(t):
                x = spmm(graph, x)
            k = torch.mm(x, x.t())
            w = (torch.abs(k) + torch.abs(k.t())) / 2
            clustering = SpectralClustering(
                n_clusters=self.num_clusters, assign_labels="discretize", random_state=0
            ).fit(w.detach().cpu())
            clusters = clustering.labels_
            intra = self.compute_intra(x.cpu().numpy(), clusters)
            print("iter #%d, intra = %.4lf" % (t, intra))
            if intra > pre_intra:
                features_matrix = pre_feat
                return features_matrix
            pre_intra = intra
            pre_feat = w
        features_matrix = w
        return features_matrix.cpu()

    def compute_intra(self, x, clusters):
        num_nodes = x.shape[0]
        intra = np.zeros(self.num_clusters)
        num_per_cluster = np.zeros(self.num_clusters)
        for i in range(num_nodes):
            for j in range(i + 1, num_nodes):
                if clusters[i] == clusters[j]:
                    intra[clusters[i]] += np.sum((x[i] - x[j]) ** 2) ** 0.5
                    num_per_cluster[clusters[i]] += 1
        intra = np.array(list(filter(lambda x: x > 0, intra)))
        num_per_cluster = np.array(list(filter(lambda x: x > 0, num_per_cluster)))
        return np.mean(intra / num_per_cluster)



import os
import logging

import torch
import torch.multiprocessing as mp

from typing import List, Optional
from cogdl.wrappers import ModelWrapper

log = logging.getLogger(__name__)


class TrainingController(object):
    def __init__(self, device_ids: Optional[List[int]], dist: str = "ddp", backend: str = "nccl"):
        self.device_ids = device_ids
        self.backend = backend

    def init_controller(self):
        if self.backend == "ddp":
            pass
        elif self.backend == "dp":
            pass
        else:
            raise NotImplementedError

    def setup(self, model_w: ModelWrapper):
        mp.spawn(self.new_process, args=(), nprocs=2)

    def new_process(self) -> None:
        pass

    def init_ddp(self, global_rank: Optional[int], world_size: Optional[int]) -> None:
        # TODO: this code is duplicated in DDP and DDPSpawn, make this a function
        global_rank = global_rank if global_rank is not None else self.global_rank()
        world_size = world_size if world_size is not None else self.world_size()
        os.environ["MASTER_ADDR"] = self.master_address()
        os.environ["MASTER_PORT"] = str(self.master_port())

        if not torch.distributed.is_initialized():
            log.info(f"initializing ddp: GLOBAL_RANK: {global_rank}, MEMBER: {global_rank + 1}/{world_size}")
            torch.distributed.init_process_group(self.backend, rank=global_rank, world_size=world_size)

    @property
    def torch_distributed_backend(self):
        torch_backend = os.getenv("PL_TORCH_DISTRIBUTED_BACKEND")
        if torch_backend is None:
            torch_backend = "nccl" if self.on_gpu else "gloo"
        return torch_backend

    def local_rank(self) -> int:
        return int(os.environ["LOCAL_RANK"])

    def master_address(self) -> str:
        if "MASTER_ADDR" not in os.environ:
            # rank_zero_warn("MASTER_ADDR environment variable is not defined. Set as localhost")
            os.environ["MASTER_ADDR"] = "127.0.0.1"
        log.debug(f"MASTER_ADDR: {os.environ['MASTER_ADDR']}")
        master_address = os.environ.get("MASTER_ADDR")
        return master_address

    def master_port(self) -> int:
        if "MASTER_PORT" not in os.environ:
            # rank_zero_warn("MASTER_PORT environment variable is not defined. Set as 12910")
            os.environ["MASTER_PORT"] = "12910"
        log.debug(f"MASTER_PORT: {os.environ['MASTER_PORT']}")

        port = int(os.environ.get("MASTER_PORT"))
        return port



import glob
import os
import os.path as osp
import shutil
import zipfile

import numpy as np
import torch
import torch.nn.functional as F

from cogdl.data.dataset import MultiGraphDataset
from cogdl.data import Graph
from cogdl.utils import download_url


def normalize_feature(data):
    x_sum = torch.sum(data.x, dim=1)
    x_rev = x_sum.pow(-1).flatten()
    x_rev[torch.isnan(x_rev)] = 0.0
    x_rev[torch.isinf(x_rev)] = 0.0
    data.x = data.x * x_rev.unsqueeze(-1).expand_as(data.x)
    return data


def parse_txt_array(src, sep=None, start=0, end=None, dtype=None, device=None):
    src = [[float(x) for x in line.split(sep)[start:end]] for line in src]
    src = torch.tensor(src, dtype=dtype).squeeze()
    return src


def read_txt_array(path, sep=None, start=0, end=None, dtype=None, device=None):
    with open(path, "r") as f:
        src = f.read().split("\n")[:-1]
    return parse_txt_array(src, sep, start, end, dtype, device)


def read_file(folder, prefix, name, dtype=None):
    path = osp.join(folder, "{}_{}.txt".format(prefix, name))
    return read_txt_array(path, sep=",", dtype=dtype)


def cat(seq):
    seq = [item for item in seq if item is not None]
    seq = [item.unsqueeze(-1) if item.dim() == 1 else item for item in seq]
    return torch.cat(seq, dim=-1) if len(seq) > 0 else None


def _split(edge_index, batch, x=None, y=None, edge_attr=None):
    node_slice = np.bincount(batch).tolist()
    row, _ = edge_index
    edge_slice = np.bincount(batch[row]).tolist()
    if edge_attr is not None:
        edge_attr = edge_attr.split(edge_slice)
    edge_index_t = edge_index.T.split(edge_slice)
    if x is not None:
        x = x.split(node_slice)
        num_nodes = [i.shape[0] for i in x]
        num_nodes_cum = np.cumsum(num_nodes).tolist()
    else:
        num_nodes_cum = [edge.max().item() + 1 for edge in edge_index_t]

    num_nodes_cum = [0] + num_nodes_cum
    if edge_index_t[-1].min() > 0:
        edge_index_t = [edge_index_t[i] - num_nodes_cum[i] for i in range(len(edge_index_t))]
    data = []
    for i in range(len(node_slice)):
        g = Graph(edge_index=edge_index_t[i].T)
        if x is not None:
            g.x = x[i]
        if y is not None:
            g.y = y[i].view(1)
        if edge_attr is not None:
            g.edge_attr = edge_attr[i]
        data.append(g)
    return data


def segment(src, indptr):
    out_list = []
    for i in range(indptr.size(-1) - 1):
        indexptr = torch.arange(indptr[..., i].item(), indptr[..., i + 1].item(), dtype=torch.int64)
        src_data = src.index_select(indptr.dim() - 1, indexptr)
        out = torch.sum(src_data, dim=indptr.dim() - 1, keepdim=True)
        out_list.append(out)
    return torch.cat(out_list, dim=indptr.dim() - 1)


def coalesce(index, value, m, n):
    row = index[0]
    col = index[1]

    idx = col.new_zeros(col.numel() + 1)
    idx[1:] = row
    idx[1:] *= n
    idx[1:] += col
    if (idx[1:] < idx[:-1]).any():
        perm = idx[1:].argsort()
        row = row[perm]
        col = col[perm]
        if value is not None:
            value = value[perm]

    idx = col.new_full((col.numel() + 1,), -1)
    idx[1:] = n * row + col
    mask = idx[1:] > idx[:-1]

    if mask.all():  # Skip if indices are already coalesced.
        return torch.stack([row, col], dim=0), value

    row = row[mask]
    col = col[mask]

    if value is not None:
        ptr = mask.nonzero().flatten()
        ptr = torch.cat([ptr, ptr.new_full((1,), value.size(0))])
        value = segment(value, ptr)
        value = value[0] if isinstance(value, tuple) else value

    return torch.stack([row, col], dim=0), value


def read_tu_data(folder, prefix):
    files = glob.glob(osp.join(folder, "{}_*.txt".format(prefix)))
    names = [f.split(os.sep)[-1][len(prefix) + 1 : -4] for f in files]

    edge_index = read_file(folder, prefix, "A", torch.long).t() - 1
    batch = read_file(folder, prefix, "graph_indicator", torch.long) - 1

    node_attributes = node_labels = None
    if "node_attributes" in names:
        node_attributes = read_file(folder, prefix, "node_attributes")
    if "node_labels" in names:
        node_labels = read_file(folder, prefix, "node_labels", torch.long)
        if node_labels.dim() == 1:
            node_labels = node_labels.unsqueeze(-1)
        node_labels = node_labels - node_labels.min(dim=0)[0]
        node_labels = node_labels.unbind(dim=-1)
        node_labels = [F.one_hot(x, num_classes=-1) for x in node_labels]
        node_labels = torch.cat(node_labels, dim=-1).to(torch.float)
    x = cat([node_attributes, node_labels])

    edge_attributes, edge_labels = None, None
    if "edge_attributes" in names:
        edge_attributes = read_file(folder, prefix, "edge_attributes")
    if "edge_labels" in names:
        edge_labels = read_file(folder, prefix, "edge_labels", torch.long)
        if edge_labels.dim() == 1:
            edge_labels = edge_labels.unsqueeze(-1)
        edge_labels = edge_labels - edge_labels.min(dim=0)[0]
        edge_labels = edge_labels.unbind(dim=-1)
        edge_labels = [F.one_hot(e, num_classes=-1) for e in edge_labels]
        edge_labels = torch.cat(edge_labels, dim=-1).to(torch.float)
    edge_attr = cat([edge_attributes, edge_labels])

    y = None
    if "graph_attributes" in names:  # Regression problem.
        y = read_file(folder, prefix, "graph_attributes")
    elif "graph_labels" in names:  # Classification problem.
        y = read_file(folder, prefix, "graph_labels", torch.long)
        _, y = y.unique(sorted=True, return_inverse=True)

    num_nodes = edge_index.max().item() + 1 if x is None else x.size(0)

    mask = edge_index[0] != edge_index[1]
    edge_index = edge_index[:, mask]
    if edge_attr is not None:
        edge_attr = edge_attr[mask]

    edge_index, edge_attr = coalesce(edge_index, edge_attr, num_nodes, num_nodes)
    if x is not None:
        x = x[:, num_node_attributes(x) :]
    if edge_attr is not None:
        edge_attr = edge_attr[:, : num_edge_attributes(edge_attr)]

    graphs = _split(edge_index, batch=batch, x=x, y=y, edge_attr=edge_attr)
    return graphs, y


def num_node_labels(x=None):
    if x is None:
        return 0
    for i in range(x.size(1)):
        _x = x[:, i:]
        if ((_x == 0) | (_x == 1)).all() and (_x.sum(dim=1) == 1).all():
            return x.size(1) - i
    return 0


def num_node_attributes(x=None):
    if x is None:
        return 0
    return x.size(1) - num_node_labels(x)


def num_edge_labels(edge_attr=None):
    if edge_attr is None:
        return 0
    for i in range(edge_attr.size(1)):
        if edge_attr[:, i:].sum() == edge_attr.size(0):
            return edge_attr.size(1) - i
    return 0


def num_edge_attributes(edge_attr=None):
    if edge_attr is None:
        return 0
    return edge_attr.size(1) - num_edge_labels(edge_attr)


class TUDataset(MultiGraphDataset):
    url = "https://www.chrsmrrs.com/graphkerneldatasets"

    def __init__(self, root, name):
        self.name = name
        super(TUDataset, self).__init__(root)
        # self.data = torch.load(self.processed_paths[0])

        # if self.data[0].x is not None:
        #     num_node_attributes = self.num_node_attributes
        #     self.data.x = self.data.x[:, num_node_attributes:]
        # if self.data.edge_attr is not None:
        #     num_edge_attributes = self.num_edge_attributes
        #     self.data.edge_attr = self.data.edge_attr[:, num_edge_attributes:]

        self.data, self.y = torch.load(self.processed_paths[0])

    @property
    def raw_file_names(self):
        names = ["A", "graph_indicator"]
        return ["{}_{}.txt".format(self.name, name) for name in names]

    @property
    def processed_file_names(self):
        return "data.pt"

    def download(self):
        url = self.url
        folder = osp.join(self.root)
        path = download_url("{}/{}.zip".format(url, self.name), folder)
        with zipfile.ZipFile(path, "r") as f:
            f.extractall(folder)
        os.unlink(path)
        shutil.rmtree(self.raw_dir)
        os.rename(osp.join(folder, self.name), self.raw_dir)

    def process(self):
        data = read_tu_data(self.raw_dir, self.name)
        torch.save(data, self.processed_paths[0])

    @property
    def num_classes(self):
        r"""The number of classes in the dataset."""
        return self.y.max().item() + 1 if self.y.dim() == 1 else self.y.size(1)

    def __len__(self):
        return len(self.data)


class MUTAGDataset(TUDataset):
    def __init__(self, data_path="data"):
        dataset = "MUTAG"
        path = osp.join(data_path, dataset)
        if not osp.exists(path):
            TUDataset(path, name=dataset)
        super(MUTAGDataset, self).__init__(path, name=dataset)


class ImdbBinaryDataset(TUDataset):
    def __init__(self, data_path="data"):
        dataset = "IMDB-BINARY"
        path = osp.join(data_path, dataset)
        if not osp.exists(path):
            TUDataset(path, name=dataset)
        super(ImdbBinaryDataset, self).__init__(path, name=dataset)


class ImdbMultiDataset(TUDataset):
    def __init__(self, data_path="data"):
        dataset = "IMDB-MULTI"
        path = osp.join(data_path, dataset)
        if not osp.exists(path):
            TUDataset(path, name=dataset)
        super(ImdbMultiDataset, self).__init__(path, name=dataset)


class CollabDataset(TUDataset):
    def __init__(self, data_path="data"):
        dataset = "COLLAB"
        path = osp.join(data_path, dataset)
        if not osp.exists(path):
            TUDataset(path, name=dataset)
        super(CollabDataset, self).__init__(path, name=dataset)


class ProteinsDataset(TUDataset):
    def __init__(self, data_path="data"):
        dataset = "PROTEINS"
        path = osp.join(data_path, dataset)
        if not osp.exists(path):
            TUDataset(path, name=dataset)
        super(ProteinsDataset, self).__init__(path, name=dataset)


class RedditBinary(TUDataset):
    def __init__(self, data_path="data"):
        dataset = "REDDIT-BINARY"
        path = osp.join(data_path, dataset)
        if not osp.exists(path):
            TUDataset(path, name=dataset)
        super(RedditBinary, self).__init__(path, name=dataset)


class RedditMulti5K(TUDataset):
    def __init__(self, data_path="data"):
        dataset = "REDDIT-MULTI-5K"
        path = osp.join(data_path, dataset)
        if not osp.exists(path):
            TUDataset(path, name=dataset)
        super(RedditMulti5K, self).__init__(path, name=dataset)


class RedditMulti12K(TUDataset):
    def __init__(self, data_path="data"):
        dataset = "REDDIT-MULTI-12K"
        path = osp.join(data_path, dataset)
        if not osp.exists(path):
            TUDataset(path, name=dataset)
        super(RedditMulti12K, self).__init__(path, name=dataset)


class PTCMRDataset(TUDataset):
    def __init__(self, data_path="data"):
        dataset = "PTC_MR"
        path = osp.join(data_path, dataset)
        if not osp.exists(path):
            TUDataset(path, name=dataset)
        super(PTCMRDataset, self).__init__(path, name=dataset)


class NCI1Dataset(TUDataset):
    def __init__(self, data_path="data"):
        dataset = "NCI1"
        path = osp.join(data_path, dataset)
        if not osp.exists(path):
            TUDataset(path, name=dataset)
        super(NCI1Dataset, self).__init__(path, name=dataset)


class NCI109Dataset(TUDataset):
    def __init__(self, data_path="data"):
        dataset = "NCI109"
        path = osp.join(data_path, dataset)
        if not osp.exists(path):
            TUDataset(path, name=dataset)
        super(NCI109Dataset, self).__init__(path, name=dataset)


class ENZYMES(TUDataset):
    def __init__(self, data_path="data"):
        dataset = "ENZYMES"
        path = osp.join(data_path, dataset)
        if not osp.exists(path):
            TUDataset(path, name=dataset)
        super(ENZYMES, self).__init__(path, name=dataset)

    def __getitem__(self, idx):
        if isinstance(idx, int):
            data = self.get(self.indices()[idx])
            data = data if self.transform is None else self.transform(data)
            edge_nodes = data.edge_index.max() + 1
            if edge_nodes < data.x.size(0):
                data.x = data.x[:edge_nodes]
            return data
        else:
            return self.index_select(idx)



import copy
import random
import numpy as np
import torch

from .. import ModelWrapper


class GCNMixModelWrapper(ModelWrapper):
    """
    GCNMixModelWrapper calls `forward_aux` in model
    `forward_aux` is similar to `forward` but ignores `spmm` operation.
    """

    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--temperature", type=float, default=0.1)
        parser.add_argument("--rampup-starts", type=int, default=500)
        parser.add_argument("--rampup-ends", type=int, default=1000)
        parser.add_argument("--mixup-consistency", type=float, default=10.0)
        parser.add_argument("--ema-decay", type=float, default=0.999)
        parser.add_argument("--tau", type=float, default=1.0)
        parser.add_argument("--k", type=int, default=10)
        # fmt: on

    def __init__(
        self, model, optimizer_cfg, temperature, rampup_starts, rampup_ends, mixup_consistency, ema_decay, tau, k
    ):
        super(GCNMixModelWrapper, self).__init__()
        self.optimizer_cfg = optimizer_cfg
        self.temperature = temperature
        self.ema_decay = ema_decay
        self.tau = tau
        self.k = k

        self.model = model
        self.model_ema = copy.deepcopy(self.model)

        for p in self.model_ema.parameters():
            p.detach_()
        self.epoch = 0
        self.opt = {
            "epoch": 0,
            "final_consistency_weight": mixup_consistency,
            "rampup_starts": rampup_starts,
            "rampup_ends": rampup_ends,
        }
        self.mix_loss = torch.nn.BCELoss()
        self.mix_transform = None

    def train_step(self, subgraph):
        if self.mix_transform is None:
            if len(subgraph.y.shape) > 1:
                self.mix_transform = torch.nn.Sigmoid()
            else:
                self.mix_transform = torch.nn.Softmax(-1)
        graph = subgraph
        device = graph.x.device
        train_mask = graph.train_mask

        self.opt["epoch"] += 1

        rand_n = random.randint(0, 1)
        if rand_n == 0:
            vector_labels = get_one_hot_label(graph.y, train_mask).to(device)
            loss = self.update_aux(graph, vector_labels, train_mask)
        else:
            loss = self.update_soft(graph)

        alpha = min(1 - 1 / (self.epoch + 1), self.ema_decay)
        for ema_param, param in zip(self.model_ema.parameters(), self.model.parameters()):
            ema_param.data.mul_(alpha).add_((1 - alpha) * param.data)

        return loss

    def val_step(self, subgraph):
        graph = subgraph
        val_mask = graph.val_mask
        pred = self.model_ema(graph)
        loss = self.default_loss_fn(pred[val_mask], graph.y[val_mask])

        metric = self.evaluate(pred[val_mask], graph.y[val_mask], metric="auto")

        self.note("val_loss", loss.item())
        self.note("val_metric", metric)

    def test_step(self, subgraph):
        test_mask = subgraph.test_mask
        pred = self.model_ema(subgraph)
        loss = self.default_loss_fn(pred[test_mask], subgraph.y[test_mask])

        metric = self.evaluate(pred[test_mask], subgraph.y[test_mask], metric="auto")

        self.note("test_loss", loss.item())
        self.note("test_metric", metric)

    def update_soft(self, graph):
        out = self.model(graph)
        train_mask = graph.train_mask
        loss_sum = self.default_loss_fn(out[train_mask], graph.y[train_mask])
        return loss_sum

    def update_aux(self, data, vector_labels, train_index):
        device = self.device
        train_unlabelled = torch.where(~data.train_mask)[0].to(device)
        temp_labels = torch.zeros(self.k, vector_labels.shape[0], vector_labels.shape[1]).to(device)
        with torch.no_grad():
            for i in range(self.k):
                temp_labels[i, :, :] = self.model(data) / self.tau

        target_labels = temp_labels.mean(dim=0)
        target_labels = sharpen(target_labels, self.temperature)
        vector_labels[train_unlabelled] = target_labels[train_unlabelled]
        sampled_unlabelled = torch.randint(0, train_unlabelled.shape[0], size=(train_index.shape[0],))
        train_unlabelled = train_unlabelled[sampled_unlabelled]

        def get_loss(index):
            # TODO: call `forward_aux` in model
            mix_logits, target = self.model.forward_aux(data.x, vector_labels, index, mix_hidden=True)
            # temp_loss = self.loss_f(F.softmax(mix_logits[index], -1), target)
            temp_loss = self.mix_loss(self.mix_transform(mix_logits[index]), target)
            return temp_loss

        sup_loss = get_loss(train_index)
        unsup_loss = get_loss(train_unlabelled)

        mixup_weight = get_current_consistency_weight(
            self.opt["final_consistency_weight"], self.opt["rampup_starts"], self.opt["rampup_ends"], self.opt["epoch"]
        )

        loss_sum = sup_loss + mixup_weight * unsup_loss
        return loss_sum

    def setup_optimizer(self):
        lr = self.optimizer_cfg["lr"]
        wd = self.optimizer_cfg["weight_decay"]
        return torch.optim.Adam(self.parameters(), lr=lr, weight_decay=wd)


def get_one_hot_label(labels, index):
    num_classes = int(torch.max(labels) + 1)
    target = torch.zeros(labels.shape[0], num_classes).to(labels.device)

    target[index, labels[index]] = 1
    return target


def sharpen(prob, temperature):
    prob = torch.pow(prob, 1.0 / temperature)
    row_sum = torch.sum(prob, dim=1).reshape(-1, 1)
    return prob / row_sum


def get_current_consistency_weight(final_consistency_weight, rampup_starts, rampup_ends, epoch):
    # Consistency ramp-up from https://arxiv.org/abs/1610.02242
    rampup_length = rampup_ends - rampup_starts
    rampup = 1.0
    epoch = epoch - rampup_starts
    if rampup_length != 0:
        current = np.clip(epoch, 0.0, rampup_length)
        phase = 1.0 - current / rampup_length
        rampup = float(np.exp(-5.0 * phase * phase))
    return final_consistency_weight * rampup



import time

import networkx as nx
import numpy as np
from sklearn import preprocessing
from tqdm import tqdm

from cogdl.utils import alias_draw, alias_setup
from .. import BaseModel


class LINE(BaseModel):
    r"""The LINE model from the `"Line: Large-scale information network embedding"
    <http://arxiv.org/abs/1503.03578>`_ paper.

    Args:
        hidden_size (int) : The dimension of node representation.
        walk_length (int) : The walk length.
        walk_num (int) : The number of walks to sample for each node.
        negative (int) : The number of nagative samples for each edge.
        batch_size (int) : The batch size of training in LINE.
        alpha (float) : The initial learning rate of SGD.
        order (int) : 1 represents perserving 1-st order proximity, 2 represents 2-nd,
        while 3 means both of them (each of them having dimension/2 node representation).
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--walk-length", type=int, default=80,
                            help="Length of walk per source. Default is 80.")
        parser.add_argument("--walk-num", type=int, default=20,
                            help="Number of walks per source. Default is 20.")
        parser.add_argument("--negative", type=int, default=5,
                            help="Number of negative node in sampling. Default is 5.")
        parser.add_argument("--batch-size", type=int, default=1000,
                            help="Batch size in SGD training process. Default is 1000.")
        parser.add_argument("--alpha", type=float, default=0.025,
                            help="Initial learning rate of SGD. Default is 0.025.")
        parser.add_argument("--order", type=int, default=3,
                            help="Order of proximity in LINE. Default is 3 for 1+2.")
        parser.add_argument("--hidden-size", type=int, default=128)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.hidden_size,
            args.walk_length,
            args.walk_num,
            args.negative,
            args.batch_size,
            args.alpha,
            args.order,
        )

    def __init__(self, dimension, walk_length, walk_num, negative, batch_size, alpha, order):
        super(LINE, self).__init__()
        self.dimension = dimension
        self.walk_length = walk_length
        self.walk_num = walk_num
        self.negative = negative
        self.batch_size = batch_size
        self.init_alpha = alpha
        self.order = order

    def forward(self, graph, return_dict=False):
        # run LINE algorithm, 1-order, 2-order or 3(1-order + 2-order)
        nx_g = graph.to_networkx()
        self.G = nx_g
        self.is_directed = nx.is_directed(self.G)
        self.num_node = nx_g.number_of_nodes()
        self.num_edge = nx_g.number_of_edges()
        self.num_sampling_edge = self.walk_length * self.walk_num * self.num_node

        node2id = dict([(node, vid) for vid, node in enumerate(nx_g.nodes())])
        self.edges = [[node2id[e[0]], node2id[e[1]]] for e in self.G.edges()]
        self.edges_prob = np.asarray([nx_g[u][v].get("weight", 1.0) for u, v in nx_g.edges()])
        self.edges_prob /= np.sum(self.edges_prob)
        self.edges_table, self.edges_prob = alias_setup(self.edges_prob)

        degree_weight = np.asarray([0] * self.num_node)
        for u, v in nx_g.edges():
            degree_weight[node2id[u]] += nx_g[u][v].get("weight", 1.0)
            if not self.is_directed:
                degree_weight[node2id[v]] += nx_g[u][v].get("weight", 1.0)
        self.node_prob = np.power(degree_weight, 0.75)
        self.node_prob /= np.sum(self.node_prob)
        self.node_table, self.node_prob = alias_setup(self.node_prob)

        if self.order == 3:
            self.dimension = int(self.dimension / 2)
        if self.order == 1 or self.order == 3:
            print("train line with 1-order")
            print(type(self.dimension))
            self.emb_vertex = (np.random.random((self.num_node, self.dimension)) - 0.5) / self.dimension
            self._train_line(order=1)
            embedding1 = preprocessing.normalize(self.emb_vertex, "l2")

        if self.order == 2 or self.order == 3:
            print("train line with 2-order")
            self.emb_vertex = (np.random.random((self.num_node, self.dimension)) - 0.5) / self.dimension
            self.emb_context = self.emb_vertex
            self._train_line(order=2)
            embedding2 = preprocessing.normalize(self.emb_vertex, "l2")

        if self.order == 1:
            embeddings = embedding1
        elif self.order == 2:
            embeddings = embedding2
        else:
            print("concatenate two embedding...")
            embeddings = np.hstack((embedding1, embedding2))

        if return_dict:
            features_matrix = dict()
            for vid, node in enumerate(nx_g.nodes()):
                features_matrix[node] = embeddings[vid]
        else:
            features_matrix = np.zeros((graph.num_nodes, embeddings.shape[1]))
            nx_nodes = nx_g.nodes()
            features_matrix[nx_nodes] = embeddings[np.arange(graph.num_nodes)]
        return features_matrix

    def _update(self, vec_u, vec_v, vec_error, label):
        # update vetex embedding and vec_error
        f = 1 / (1 + np.exp(-np.sum(vec_u * vec_v, axis=1)))
        g = (self.alpha * (label - f)).reshape((len(label), 1))
        vec_error += g * vec_v
        vec_v += g * vec_u

    def _train_line(self, order):
        # train Line model with order
        self.alpha = self.init_alpha
        batch_size = self.batch_size
        t0 = time.time()
        num_batch = int(self.num_sampling_edge / batch_size)
        epoch_iter = tqdm(range(num_batch))
        for b in epoch_iter:
            if b % 100 == 0:
                epoch_iter.set_description(
                    f"Progress: {b *1.0/num_batch * 100:.4f}%, alpha: {self.alpha:.6f}, time: {time.time() - t0:.4f}"
                )
                self.alpha = self.init_alpha * max((1 - b * 1.0 / num_batch), 0.0001)
            u, v = [0] * batch_size, [0] * batch_size
            for i in range(batch_size):
                edge_id = alias_draw(self.edges_table, self.edges_prob)
                u[i], v[i] = self.edges[edge_id]
                if not self.is_directed and np.random.rand() > 0.5:
                    v[i], u[i] = self.edges[edge_id]

            vec_error = np.zeros((batch_size, self.dimension))
            label, target = np.asarray([1 for i in range(batch_size)]), np.asarray(v)
            for j in range(1 + self.negative):
                if j != 0:
                    label = np.asarray([0 for i in range(batch_size)])
                    for i in range(batch_size):
                        target[i] = alias_draw(self.node_table, self.node_prob)
                if order == 1:
                    self._update(self.emb_vertex[u], self.emb_vertex[target], vec_error, label)
                else:
                    self._update(self.emb_vertex[u], self.emb_context[target], vec_error, label)
            self.emb_vertex[u] += vec_error



import torch
import torch.nn.functional as F

from cogdl.wrappers.model_wrapper.node_classification.node_classification_mw import NodeClfModelWrapper


class GrandModelWrapper(NodeClfModelWrapper):
    """
    sample : int
        Number of augmentations for consistency loss
    temperature : float
        Temperature to sharpen predictions.
    lmbda : float
         Proportion of consistency loss of unlabelled data
    """

    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--temperature", type=float, default=0.5)
        parser.add_argument("--lmbda", type=float, default=0.5)
        parser.add_argument("--sample", type=int, default=2)
        # fmt: on

    def __init__(self, model, optimizer_cfg, sample=2, temperature=0.5, lmbda=0.5):
        super(GrandModelWrapper, self).__init__(model, optimizer_cfg)
        self.sample = sample
        self.temperature = temperature
        self.lmbda = lmbda

    def train_step(self, batch):
        graph = batch
        output_list = []
        for i in range(self.sample):
            output_list.append(self.model(graph))
        loss_train = 0.0
        for output in output_list:
            loss_train += self.default_loss_fn(output[graph.train_mask], graph.y[graph.train_mask])
        loss_train = loss_train / self.sample

        if len(graph.y.shape) > 1:
            output_list = [torch.sigmoid(x) for x in output_list]
        else:
            output_list = [F.log_softmax(x, dim=-1) for x in output_list]
        loss_consis = self.consistency_loss(output_list, graph.train_mask)

        return loss_train + loss_consis

    def consistency_loss(self, logps, train_mask):
        temp = self.temperature
        ps = [torch.exp(p)[~train_mask] for p in logps]
        sum_p = 0.0
        for p in ps:
            sum_p = sum_p + p
        avg_p = sum_p / len(ps)
        sharp_p = (torch.pow(avg_p, 1.0 / temp) / torch.sum(torch.pow(avg_p, 1.0 / temp), dim=1, keepdim=True)).detach()
        loss = 0.0
        for p in ps:
            loss += torch.mean((p - sharp_p).pow(2).sum(1))
        loss = loss / len(ps)

        return self.lmbda * loss



from tensorboardX import SummaryWriter

from . import Logger


class TBLogger(Logger):
    def __init__(self, log_path):
        super(TBLogger, self).__init__(log_path)
        self.last_step = 0

    def start(self):
        self.writer = SummaryWriter(logdir=self.log_path)

    def note(self, metrics, step=None):
        if not hasattr(self, "writer"):
            self.start()
        if step is None:
            step = self.last_step
        for key, value in metrics.items():
            self.writer.add_scalar(key, value, step)
        self.last_step = step

    def finish(self):
        self.writer.close()



from .embedding_link_prediction_dw import EmbeddingLinkPredictionDataWrapper
from .gnn_kg_link_prediction_dw import GNNKGLinkPredictionDataWrapper
from .gnn_link_prediction_dw import GNNLinkPredictionDataWrapper



import numpy as np
import networkx as nx
import scipy.sparse as sp
from sklearn import preprocessing
from .. import BaseModel


class GraRep(BaseModel):
    r"""The GraRep model from the `"Grarep: Learning graph representations with global structural information"
    <http://dl.acm.org/citation.cfm?doid=2806416.2806512>`_ paper.

    Args:
        hidden_size (int) : The dimension of node representation.
        step (int) : The maximum order of transitition probability.
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--step", type=int, default=5,
                            help="Number of matrix step in GraRep. Default is 5.")
        parser.add_argument("--hidden-size", type=int, default=128)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(args.hidden_size, args.step)

    def __init__(self, dimension, step):
        super(GraRep, self).__init__()
        self.dimension = dimension
        self.step = step

    def forward(self, graph, return_dict=False):
        self.G = graph.to_networkx()
        self.num_node = self.G.number_of_nodes()
        A = np.asarray(nx.adjacency_matrix(self.G).todense(), dtype=float)
        A = preprocessing.normalize(A, "l1")

        log_beta = np.log(1.0 / self.num_node)
        A_list = [A]
        T_list = [sum(A).tolist()]
        temp = A
        # calculate A^1, A^2, ... , A^step, respectively
        for i in range(self.step - 1):
            temp = temp.dot(A)
            A_list.append(A)
            T_list.append(sum(temp).tolist())

        final_emb = np.zeros((self.num_node, 1))
        for k in range(self.step):
            for j in range(A.shape[1]):
                A_list[k][:, j] = np.log(A_list[k][:, j] / T_list[k][j] + 1e-20) - log_beta
                for i in range(A.shape[0]):
                    A_list[k][i, j] = max(A_list[k][i, j], 0)
            # concatenate all k-step representations
            if k == 0:
                dimension = self.dimension - int(self.dimension / self.step) * (self.step - 1)
                final_emb = self._get_embedding(A_list[k], dimension)
            else:
                W = self._get_embedding(A_list[k], self.dimension / self.step)
                final_emb = np.hstack((final_emb, W))

        embeddings = final_emb
        if return_dict:
            features_matrix = dict()
            for vid, node in enumerate(self.G.nodes()):
                features_matrix[node] = embeddings[vid]
        else:
            features_matrix = np.zeros((graph.num_nodes, embeddings.shape[1]))
            nx_nodes = self.G.nodes()
            features_matrix[nx_nodes] = embeddings[np.arange(graph.num_nodes)]
        return features_matrix

    def _get_embedding(self, matrix, dimension):
        # get embedding from svd and process normalization for ut
        ut, s, _ = sp.linalg.svds(matrix, int(dimension))
        emb_matrix = ut * np.sqrt(s)
        emb_matrix = preprocessing.normalize(emb_matrix, "l2")
        return emb_matrix



import math

import torch
import torch.nn as nn


class MemoryMoCo(nn.Module):
    """Fixed-size queue with momentum encoder"""

    def __init__(self, inputSize, outputSize, K, T=0.07, use_softmax=False):
        super(MemoryMoCo, self).__init__()
        self.outputSize = outputSize  # None
        self.inputSize = inputSize
        self.queueSize = K
        self.T = T
        self.index = 0
        self.use_softmax = use_softmax

        self.register_buffer("params", torch.tensor([-1]))
        stdv = 1.0 / math.sqrt(inputSize / 3)
        self.register_buffer("memory", torch.rand(self.queueSize, inputSize).mul_(2 * stdv).add_(-stdv))
        print("using queue shape: ({},{})".format(self.queueSize, inputSize))

    def forward(self, q, k):
        batchSize = q.shape[0]
        k = k.detach()

        Z = self.params[0].item()

        # pos logit
        l_pos = torch.bmm(q.view(batchSize, 1, -1), k.view(batchSize, -1, 1))
        l_pos = l_pos.view(batchSize, 1)
        # neg logit
        queue = self.memory.clone()
        l_neg = torch.mm(queue.detach(), q.transpose(1, 0))
        l_neg = l_neg.transpose(0, 1)

        out = torch.cat((l_pos, l_neg), dim=1)

        if self.use_softmax:
            out = torch.div(out, self.T)
            out = out.squeeze().contiguous()
        else:
            out = torch.exp(torch.div(out, self.T))
            if Z < 0:
                self.params[0] = out.mean() * self.outputSize
                Z = self.params[0].clone().detach().item()
                print("normalization constant Z is set to {:.1f}".format(Z))
            # compute the out
            out = torch.div(out, Z).squeeze().contiguous()

        # # update memory
        with torch.no_grad():
            out_ids = torch.arange(batchSize, device=out.device)
            out_ids += self.index
            out_ids = torch.fmod(out_ids, self.queueSize)
            out_ids = out_ids.long()
            self.memory.index_copy_(0, out_ids, k)
            self.index = (self.index + batchSize) % self.queueSize

        return out


class NCESoftmaxLoss(nn.Module):
    """Softmax cross-entropy loss (a.k.a., info-NCE loss in CPC paper)"""

    def __init__(self):
        super(NCESoftmaxLoss, self).__init__()
        self.criterion = nn.CrossEntropyLoss()

    def forward(self, x):
        bsz = x.shape[0]
        x = x.squeeze()
        label = torch.zeros([bsz], device=x.device).long()
        loss = self.criterion(x, label)
        return loss


def moment_update(model, model_ema, m):
    """model_ema = m * model_ema + (1 - m) model"""
    for p1, p2 in zip(model.parameters(), model_ema.parameters()):
        # p2.data.mul_(m).add_(1 - m, p1.detach().data)
        p2.data.mul_(m).add_(p1.detach().data * (1 - m))



import torch
import torch.nn as nn


class BaseLayer(nn.Module):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def forward(self, graph, x):
        m = self.message(x[graph.edge_index[0]])
        return self.aggregate(graph, m)

    def message(self, x):
        return x

    def aggregate(self, graph, x):
        result = torch.zeros(graph.num_nodes, x.shape[1], dtype=x.dtype).to(x.device)
        result.scatter_add_(0, graph.edge_index[1].unsqueeze(1).expand(-1, x.shape[1]), x)
        return result



import os
import numpy as np
import time
from typing import Optional
import torch


class EmbeddingTrainer(object):
    def __init__(
        self,
        save_emb_path: Optional[str] = None,
        load_emb_path: Optional[str] = None,
    ):
        self.save_emb_path = save_emb_path
        self.load_emb_path = load_emb_path
        self.default_emb_dir = "./embeddings"

    def run(self, model_w, dataset_w):
        self.prepare_data_wrapper(dataset_w)
        if self.load_emb_path is not None:
            print(f"Loading embeddings from {self.load_emb_path} ...")
            embedding = np.load(self.load_emb_path)
            return self.test(model_w, dataset_w, embedding)

        if self.save_emb_path is None:
            cur_time = time.strftime("%m-%d_%H.%M.%S", time.localtime())
            name = f"{model_w.wrapped_model.__class__.__name__}_{cur_time}.emb"
            self.save_emb_path = os.path.join(self.default_emb_dir, name)
            os.makedirs(self.default_emb_dir, exist_ok=True)
        embeddings = self.train(model_w, dataset_w)
        self.save_embedding(embeddings)
        return self.test(model_w, dataset_w, embeddings)

    def prepare_data_wrapper(self, dataset_w):
        dataset_w.pre_transform()
        dataset_w.prepare_training_data()
        dataset_w.prepare_val_data()
        dataset_w.prepare_test_data()

    def train(self, model_w, dataset_w):
        dataset_w.pre_transform()
        train_data = dataset_w.on_train_wrapper()
        embeddings = []
        for batch in train_data:
            embeddings.append(model_w.train_step(batch))
        assert len(embeddings) == 1
        embeddings = embeddings[0]
        return embeddings

    def test(self, model_w, dataset_w, embeddings):
        labels = next(dataset_w.on_test_wrapper())
        if torch.is_tensor(labels):
            labels = labels.cpu().numpy()
        result = model_w.test_step((embeddings, labels))
        return result

    def save_embedding(self, embeddings):
        np.save(self.save_emb_path, embeddings)



import math
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from cogdl.layers import GATLayer, SELayer, GCNLayer, GCNIILayer
from cogdl.models import BaseModel
from cogdl.utils import spmm


def gcn_model(in_feats, hidden_size, num_layers, out_feats, dropout, residual, norm, activation):
    shapes = [in_feats] + [hidden_size] * (num_layers - 1) + [out_feats]

    return nn.ModuleList(
        [
            GCNLayer(
                shapes[i],
                shapes[i + 1],
                dropout=dropout if i != num_layers - 1 else 0,
                residual=residual if i != num_layers - 1 else None,
                norm=norm if i != num_layers - 1 else None,
                activation=activation if i != num_layers - 1 else None,
            )
            for i in range(num_layers)
        ]
    )


def gat_model(
    in_feats, hidden_size, out_feats, nhead, attn_drop, alpha, residual, norm, num_layers, dropout, last_nhead
):
    layers = nn.ModuleList()
    layers.append(
        GATLayer(in_feats, hidden_size, nhead=nhead, attn_drop=attn_drop, alpha=alpha, residual=residual, norm=norm)
    )
    if num_layers != 1:
        layers.append(nn.ELU())
    for i in range(num_layers - 2):
        if dropout > 0.0:
            layers.append(nn.Dropout(dropout))
        layers.append(
            GATLayer(
                hidden_size * nhead,
                hidden_size,
                nhead=nhead,
                attn_drop=attn_drop,
                alpha=alpha,
                residual=residual,
                norm=norm,
            )
        )
        layers.append(nn.ELU())

    if dropout > 0.0:
        layers.append(nn.Dropout(p=dropout))
    layers.append(
        GATLayer(
            hidden_size * nhead,
            out_feats,
            attn_drop=attn_drop,
            alpha=alpha,
            nhead=last_nhead,
            residual=False,
        )
    )

    return layers


def grand_model(in_feats, hidden_size, out_feats, dropout, dropout2, norm):
    layers = nn.ModuleList()
    if norm == "batchnorm":
        layers.append(nn.BatchNorm1d(in_feats))
    layers.append(nn.Dropout(p=dropout))  # dropout=inputdropout
    layers.append(nn.Linear(in_feats, hidden_size))
    layers.append(nn.ReLU())
    if norm == "batchnorm":
        layers.append(nn.BatchNorm1d(hidden_size))
    layers.append(nn.Dropout(p=dropout2))  # dropout2
    layers.append(nn.Linear(hidden_size, out_feats))

    return layers


def gcnii_model(in_feats, hidden_size, out_feats, dropout, num_layers, alpha, lmbda, residual):
    layers = nn.ModuleList()
    layers.append(nn.Dropout(p=dropout))
    layers.append(nn.Linear(in_feats, hidden_size))
    layers.append(nn.ReLU())
    for i in range(num_layers):
        layers.append(nn.Dropout(p=dropout))
        layers.append(GCNIILayer(hidden_size, alpha, math.log(lmbda / (i + 1) + 1), residual))
        layers.append(nn.ReLU())
    layers.append(nn.Dropout(p=dropout))
    layers.append(nn.Linear(hidden_size, out_feats))

    return layers


def drgat_model(num_features, hidden_size, num_classes, dropout, num_heads):
    layers = nn.ModuleList()
    layers.append(nn.Dropout(p=dropout))
    layers.append(SELayer(num_features, se_channels=int(np.sqrt(num_features))))
    layers.append(GATLayer(num_features, hidden_size, nhead=num_heads, attn_drop=dropout))
    layers.append(nn.ELU())
    layers.append(nn.Dropout(p=dropout))
    layers.append(SELayer(hidden_size * num_heads, se_channels=int(np.sqrt(hidden_size * num_heads))))
    layers.append(GATLayer(hidden_size * num_heads, num_classes, nhead=1, attn_drop=dropout))
    layers.append(nn.ELU())

    return layers


class AutoGNN(BaseModel):
    """
    Args
    """

    @staticmethod
    def add_args(parser):
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--hidden-size", type=int, default=8)
        parser.add_argument("--layers-type", type=str, default="gcn")
        parser.add_argument("--num-layers", type=int, default=2)
        parser.add_argument("--dropout", type=float, default=0.6)
        parser.add_argument("--norm", type=str, default=None)
        parser.add_argument("--residual", action="store_true")
        parser.add_argument("--activation", type=str, default="relu")
        parser.add_argument("--attn-drop", type=float, default=0.5)
        parser.add_argument("--alpha", type=float, default=0.2)
        parser.add_argument("--nhead", type=int, default=8)
        parser.add_argument("--last-nhead", type=int, default=1)
        parser.add_argument("--weight-decay", type=float, default=0.0)
        parser.add_argument("--dropoutn", type=float, default=0.5)

    @classmethod
    def build_model_from_args(cls, args):
        if not hasattr(args, "attn_drop"):
            args.attn_drop = 0.5
        if not hasattr(args, "alpha"):
            args.alpha = 0.2
        if not hasattr(args, "nhead"):
            args.nhead = 8
        if not hasattr(args, "last_nhead"):
            args.last_nhead = 1
        if not hasattr(args, "dropoutn"):
            args.dropoutn = 0.5
        return cls(
            args.num_features,
            args.hidden_size,
            args.num_classes,
            args.num_layers,
            args.layers_type,
            args.dropout,
            args.activation,
            args.norm,
            args.residual,
            args.attn_drop,
            args.alpha,
            args.nhead,
            args.last_nhead,
            args.dropoutn,
        )

    def __init__(
        self,
        in_feats,
        hidden_size,
        out_feats,
        num_layers,
        layers_type,
        dropout,
        activation=None,
        norm=None,  # reuse `use_bn`
        residual=False,
        attn_drop=0.5,  # reuse `dropnode`
        alpha=0.2,
        nhead=8,  # reuse `order`
        last_nhead=1,
        dropoutn=0.5,  # reuse `gcnii:lambda`
    ):
        super(AutoGNN, self).__init__()

        self.dropout = dropout
        self.layers_type = layers_type
        if self.layers_type == "gcn":
            self.layers = gcn_model(in_feats, hidden_size, num_layers, out_feats, dropout, residual, norm, activation)
            self.num_layers = num_layers

        elif self.layers_type == "gat":
            self.layers = gat_model(
                in_feats,
                hidden_size,
                out_feats,
                nhead,
                attn_drop,
                alpha,
                residual,
                norm,
                num_layers,
                dropout,
                last_nhead,
            )
            self.num_layers = num_layers
            self.last_nhead = last_nhead
        elif self.layers_type == "grand":
            self.layers = grand_model(in_feats, hidden_size, out_feats, dropout, dropoutn, norm)
            self.dropnode_rate = attn_drop
            self.order = nhead
        elif self.layers_type == "gcnii":
            self.layers = gcnii_model(in_feats, hidden_size, out_feats, dropout, num_layers, alpha, dropoutn, residual)
        elif self.layers_type == "drgat":
            self.layers = drgat_model(in_feats, hidden_size, out_feats, dropout, nhead)

        self.autognn_parameters = list(self.layers.parameters())

    def drop_node(self, x):
        n = x.shape[0]
        drop_rates = torch.ones(n) * self.dropnode_rate
        if self.training:
            masks = torch.bernoulli(1.0 - drop_rates).unsqueeze(1)
            x = masks.to(x.device) * x

        else:
            x = x * (1.0 - self.dropnode_rate)
        return x

    def rand_prop(self, graph, x):
        x = self.drop_node(x)

        y = x
        for i in range(self.order):
            x = spmm(graph, x).detach_()
            y.add_(x)
        return y.div_(self.order + 1.0).detach_()

    def normalize_x(self, x):
        row_sum = x.sum(1)
        row_inv = row_sum.pow_(-1)
        row_inv.masked_fill_(row_inv == float("inf"), 0)
        x = x * row_inv[:, None]
        return x

    def forward(self, graph):
        if self.layers_type == "gcn":
            graph.sym_norm()
            h = graph.x
        elif self.layers_type == "gat":
            h = graph.x
        elif self.layers_type == "grand":
            graph.sym_norm()
            x = graph.x
            x = self.normalize_x(x)
            h = self.rand_prop(graph, x)
        elif self.layers_type == "gcnii":
            graph.sym_norm()
            h = graph.x
        elif self.layers_type == "drgat":
            h = graph.x

        init_h = None
        for i, layer in enumerate(self.layers):

            if type(layer).__name__ == "GATLayer" or type(layer).__name__ == "GCNLayer":
                h = layer(graph, h)
            elif type(layer).__name__ == "GCNIILayer":
                h = layer(graph, h, init_h)
            else:
                h = layer(h)

            if i == 2:
                init_h = h
        return h



import numpy as np
import scipy.sparse as sp
import scipy.sparse.linalg as slinalg

import torch

from .node_classification_dw import FullBatchNodeClfDataWrapper


class M3SDataWrapper(FullBatchNodeClfDataWrapper):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--label-rate", type=float, default=0.2)
        parser.add_argument("--approximate", action="store_true")
        parser.add_argument("--alpha", type=float, default=0.2)
        # fmt: on

    def __init__(self, dataset, label_rate, approximate, alpha):
        super(M3SDataWrapper, self).__init__(dataset)
        self.dataset = dataset
        self.label_rate = label_rate
        self.approximate = approximate
        self.alpha = alpha

    def pre_transform(self):
        data = self.dataset.data
        num_nodes = data.num_nodes
        num_classes = data.num_classes

        data.add_remaining_self_loops()
        train_nodes = torch.where(data.train_mask)[0]
        if len(train_nodes) / num_nodes > self.label_rate:
            perm = np.random.permutation(train_nodes.shape[0])
            preserve_nnz = int(num_nodes * self.label_rate)
            preserved = train_nodes[perm[:preserve_nnz]]
            masked = train_nodes[perm[preserve_nnz:]]
            data.train_mask = torch.full((data.train_mask.shape[0],), False, dtype=torch.bool)
            data.train_mask[preserved] = True
            data.test_mask[masked] = True

        # Compute absorption probability
        row, col = data.edge_index
        A = sp.coo_matrix(
            (np.ones(row.shape[0]), (row.numpy(), col.numpy())),
            shape=(num_nodes, num_nodes),
        ).tocsr()
        D = A.sum(1).flat
        confidence = np.zeros([num_classes, num_nodes])
        confidence_ranking = np.zeros([num_classes, num_nodes], dtype=int)

        if self.approximate:
            eps = 1e-2
            for i in range(num_classes):
                q = list(torch.where(data.y == i)[0].numpy())
                q = list(filter(lambda x: data.train_mask[x], q))
                r = {idx: 1 for idx in q}
                while len(q) > 0:
                    unode = q.pop()
                    res = self.alpha / (self.alpha + D[unode]) * r[unode] if unode in r else 0
                    confidence[i][unode] += res
                    r[unode] = 0
                    for vnode in A.indices[A.indptr[unode] : A.indptr[unode + 1]]:
                        val = res / self.alpha
                        if vnode in r:
                            r[vnode] += val
                        else:
                            r[vnode] = val
                        # print(vnode, val)
                        if val > eps * D[vnode] and vnode not in q:
                            q.append(vnode)
        else:
            L = sp.diags(D, dtype=np.float32) - A
            L += self.alpha * sp.eye(L.shape[0], dtype=L.dtype)
            P = slinalg.inv(L.tocsc()).toarray().transpose()
            for i in range(num_nodes):
                if data.train_mask[i]:
                    confidence[data.y[i]] += P[i]

        # Sort nodes by confidence for each class
        for i in range(num_classes):
            confidence_ranking[i] = np.argsort(-confidence[i])
        data.confidence_ranking = confidence_ranking

        self.dataset.data = data

    def pre_stage(self, stage, model_w_out):
        self.dataset.data.store("y")
        if stage > 0:
            self.dataset.data.y = model_w_out

    def post_stage(self, stage, model_w_out):
        self.dataset.data.restore("y")

    def get_dataset(self):
        return self.dataset



import torch.nn as nn
import torch.nn.functional as F
from cogdl.layers import GCNLayer
from cogdl.utils import get_activation
from fmoe import FMoETransformerMLP

from .. import BaseModel


class CustomizedMoEPositionwiseFF(FMoETransformerMLP):
    def __init__(self, d_model, d_inner, dropout, moe_num_expert=64, moe_top_k=2):
        activation = nn.Sequential(nn.GELU(), nn.Dropout(dropout))
        super().__init__(
            num_expert=moe_num_expert, d_model=d_model, d_hidden=d_inner, top_k=moe_top_k, activation=activation
        )

        self.dropout = nn.Dropout(dropout)
        self.bn_layer = nn.BatchNorm1d(d_model)

    def forward(self, inp):
        ##### positionwise feed-forward
        core_out = super().forward(inp)
        core_out = self.dropout(core_out)

        ##### residual connection + batch normalization
        output = self.bn_layer(inp + core_out)

        return output


class GraphConvBlock(nn.Module):
    def __init__(self, conv_func, conv_params, in_feats, out_feats, dropout=0.0, residual=False):
        super(GraphConvBlock, self).__init__()

        self.graph_conv = conv_func(**conv_params, in_features=in_feats, out_features=out_feats)
        self.pos_ff = CustomizedMoEPositionwiseFF(out_feats, out_feats * 2, dropout, moe_num_expert=64, moe_top_k=2)
        self.dropout = dropout
        if residual is True:
            assert in_feats is not None
            self.res_connection = nn.Linear(in_feats, out_feats)
        else:
            self.res_connection = None

    def reset_parameters(self):
        """Reinitialize model parameters."""
        # self.graph_conv.reset_parameters()
        if self.res_connection is not None:
            self.res_connection.reset_parameters()

    def forward(self, graph, feats):
        new_feats = self.graph_conv(graph, feats)
        if self.res_connection is not None:
            res = self.res_connection
            new_feats = new_feats + res
            new_feats = F.dropout(new_feats, p=self.dropout, training=self.training)

        new_feats = self.pos_ff(new_feats)

        return new_feats


class MoEGCN(BaseModel):
    r"""The GCN model from the `"Semi-Supervised Classification with Graph Convolutional Networks"
    <https://arxiv.org/abs/1609.02907>`_ paper

    Args:
        in_features (int) : Number of input features.
        out_features (int) : Number of classes.
        hidden_size (int) : The dimension of node representation.
        dropout (float) : Dropout rate for model training.
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--num-layers", type=int, default=2)
        parser.add_argument("--hidden-size", type=int, default=64)
        parser.add_argument("--dropout", type=float, default=0.5)
        parser.add_argument("--no-residual", action="store_true")
        parser.add_argument("--norm", type=str, default="batchnorm")
        parser.add_argument("--activation", type=str, default="relu")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.hidden_size,
            args.num_classes,
            args.num_layers,
            args.dropout,
            args.activation,
            not args.no_residual,
            args.norm,
        )

    def __init__(
        self, in_feats, hidden_size, out_feats, num_layers, dropout, activation="relu", residual=True, norm=None
    ):
        super(MoEGCN, self).__init__()
        shapes = [in_feats] + [hidden_size] * num_layers
        conv_func = GCNLayer
        conv_params = {
            "dropout": dropout,
            "norm": norm,
            "residual": residual,
            "activation": activation,
        }
        self.layers = nn.ModuleList(
            [
                GraphConvBlock(
                    conv_func,
                    conv_params,
                    shapes[i],
                    shapes[i + 1],
                    dropout=dropout,
                )
                for i in range(num_layers)
            ]
        )
        self.num_layers = num_layers
        self.dropout = dropout
        self.act = get_activation(activation)
        self.final_cls = nn.Linear(hidden_size, out_feats)

    def embed(self, graph):
        graph.sym_norm()
        h = graph.x
        for i in range(self.num_layers - 1):
            h = self.layers[i](graph, h)
        return h

    def forward(self, graph):
        graph.sym_norm()
        h = graph.x
        for i in range(self.num_layers):
            h = self.layers[i](graph, h)
        h = self.final_cls(h)
        return h

    def predict(self, data):
        return self.forward(data)



import torch
import torch.nn as nn
import torch.nn.functional as F

from cogdl.data import Graph
from .. import ModelWrapper
from cogdl.wrappers.tools.wrapper_utils import evaluate_node_embeddings_using_logreg
from cogdl.utils import dropout_adj, dropout_features


class GRACEModelWrapper(ModelWrapper):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--tau", type=float, default=0.5)
        parser.add_argument("--drop-feature-rates", type=float, nargs="+", default=[0.3, 0.4])
        parser.add_argument("--drop-edge-rates", type=float, nargs="+", default=[0.2, 0.4])
        parser.add_argument("--batch-fwd", type=int, default=-1)
        parser.add_argument("--proj-hidden-size", type=int, default=128)
        # fmt: on

    def __init__(self, model, optimizer_cfg, tau, drop_feature_rates, drop_edge_rates, batch_fwd, proj_hidden_size):
        super(GRACEModelWrapper, self).__init__()
        self.tau = tau
        self.drop_feature_rates = drop_feature_rates
        self.drop_edge_rates = drop_edge_rates
        self.batch_size = batch_fwd

        self.model = model
        hidden_size = optimizer_cfg["hidden_size"]
        self.project_head = nn.Sequential(
            nn.Linear(hidden_size, proj_hidden_size), nn.ELU(), nn.Linear(proj_hidden_size, hidden_size)
        )
        self.optimizer_cfg = optimizer_cfg

    def train_step(self, subgraph):
        graph = subgraph
        z1 = self.prop(graph, graph.x, self.drop_feature_rates[0], self.drop_edge_rates[0])
        z2 = self.prop(graph, graph.x, self.drop_feature_rates[1], self.drop_edge_rates[1])

        z1 = self.project_head(z1)
        z2 = self.project_head(z2)

        if self.batch_size > 0:
            return 0.5 * (self.batched_loss(z1, z2, self.batch_size) + self.batched_loss(z2, z1, self.batch_size))
        else:
            return 0.5 * (self.contrastive_loss(z1, z2) + self.contrastive_loss(z2, z1))

    def test_step(self, graph):
        with torch.no_grad():
            pred = self.model(graph)
        y = graph.y
        result = evaluate_node_embeddings_using_logreg(pred, y, graph.train_mask, graph.test_mask)
        self.note("test_acc", result)

    def prop(
        self,
        graph: Graph,
        x: torch.Tensor,
        drop_feature_rate: float = 0.0,
        drop_edge_rate: float = 0.0,
    ):
        x = dropout_features(x, drop_feature_rate)
        with graph.local_graph():
            graph.edge_index, graph.edge_weight = dropout_adj(graph.edge_index, graph.edge_weight, drop_edge_rate)
            return self.model.forward(graph, x)

    def contrastive_loss(self, z1: torch.Tensor, z2: torch.Tensor):
        z1 = F.normalize(z1, p=2, dim=-1)
        z2 = F.normalize(z2, p=2, dim=-1)

        def score_func(emb1, emb2):
            scores = torch.matmul(emb1, emb2.t())
            scores = torch.exp(scores / self.tau)
            return scores

        intro_scores = score_func(z1, z1)
        inter_scores = score_func(z1, z2)

        _loss = -torch.log(intro_scores.diag() / (intro_scores.sum(1) - intro_scores.diag() + inter_scores.sum(1)))
        return torch.mean(_loss)

    def batched_loss(
        self,
        z1: torch.Tensor,
        z2: torch.Tensor,
        batch_size: int,
    ):
        num_nodes = z1.shape[0]
        num_batches = (num_nodes - 1) // batch_size + 1

        losses = []
        indices = torch.arange(num_nodes).to(z1.device)
        for i in range(num_batches):
            train_indices = indices[i * batch_size : (i + 1) * batch_size]
            _loss = self.contrastive_loss(z1[train_indices], z2)
            losses.append(_loss)
        return sum(losses) / len(losses)

    def setup_optimizer(self):
        cfg = self.optimizer_cfg
        return torch.optim.Adam(self.parameters(), lr=cfg["lr"], weight_decay=cfg["weight_decay"])



import os
import random
import tarfile

import numpy as np
import requests
import torch
from torch import Tensor

from cogdl.data import Graph
from cogdl.datasets import NodeDataset

base_url = "https://cloud.tsinghua.edu.cn/d/65d7c53dd8474d7091a9/files/?p=%2F"


def get_whole_mask(y, ratio: str, seed: int = 1234567):
    """split the whole dataset in proportion"""
    y_have_label_mask = y != -1
    total_node_num = len(y)
    y_index_tensor = torch.tensor(list(range(total_node_num)), dtype=int)
    masked_index = y_index_tensor[y_have_label_mask]
    while True:
        (train_mask, val_mask, test_mask) = get_order(ratio, masked_index, total_node_num, seed)
        if check_train_containing(train_mask, y):
            return (train_mask, val_mask, test_mask)
        else:
            seed += 1


def get_order(ratio: str, masked_index: Tensor, total_node_num: int, seed: int = 1234567):
    """
    return：(train_mask,val_mask,test_mask)
    """
    random.seed(seed)

    masked_node_num = len(masked_index)
    shuffle_criterion = list(range(masked_node_num))
    random.shuffle(shuffle_criterion)

    train_val_test_list = [int(i) for i in ratio.split("-")]
    tvt_sum = sum(train_val_test_list)
    tvt_ratio_list = [i / tvt_sum for i in train_val_test_list]

    train_end_index = int(tvt_ratio_list[0] * masked_node_num)
    val_end_index = train_end_index + int(tvt_ratio_list[1] * masked_node_num)

    train_mask_index = shuffle_criterion[:train_end_index]
    val_mask_index = shuffle_criterion[train_end_index:val_end_index]
    test_mask_index = shuffle_criterion[val_end_index:]

    train_mask = torch.zeros(total_node_num, dtype=torch.bool)
    train_mask[masked_index[train_mask_index]] = True
    val_mask = torch.zeros(total_node_num, dtype=torch.bool)
    val_mask[masked_index[val_mask_index]] = True
    test_mask = torch.zeros(total_node_num, dtype=torch.bool)
    test_mask[masked_index[test_mask_index]] = True

    return (train_mask, val_mask, test_mask)


def check_train_containing(train_mask, y):
    for label in y.unique():
        if label.item() == -1:
            continue
        if label.item() not in y[train_mask]:
            return False
    return True


class RD2CD(NodeDataset):
    def __init__(self, root, name):
        self.root = root
        self.name = name
        self.source_path = root + "/" + name + "/raw"
        if not os.path.exists(self.source_path):
            os.makedirs(self.source_path)
        dst_path = root + "/" + name + "/processed"
        if not os.path.exists(dst_path):
            os.makedirs(dst_path)
        self.data_path = dst_path + "/data.pt"
        if not os.path.exists(self.data_path):
            self.download()
        super(RD2CD, self).__init__(path=self.data_path, scale_feat=False)

    def download(self):
        r = requests.get(base_url + self.name + ".tgz&dl=1")
        tarfile_path = self.source_path + "/" + self.name + ".tgz"
        with open(tarfile_path, "wb") as f:
            f.write(r.content)
        with tarfile.open(tarfile_path, "r") as f:
            f.extractall(self.source_path)

    def process(self):
        numpy_x = np.load(self.source_path + "/x.npy")
        x = torch.from_numpy(numpy_x).to(torch.float)
        numpy_y = np.load(self.source_path + "/y.npy")
        y = torch.from_numpy(numpy_y).to(torch.long)
        numpy_edge_index = np.load(self.source_path + "/edge_index.npy")
        edge_index = torch.from_numpy(numpy_edge_index).to(torch.long)

        # set train/val/test mask in node_classification task
        random_seed = 14530529  # a fixed seed
        (train_mask, val_mask, test_mask) = get_whole_mask(y, "6-2-2", random_seed)
        data = Graph(x=x, edge_index=edge_index, y=y, train_mask=train_mask, val_mask=val_mask, test_mask=test_mask)
        torch.save(data, self.data_path)
        return data


class Github(RD2CD):
    def __init__(self, root="data"):
        super(Github, self).__init__(root, "Github")


class Elliptic(RD2CD):
    def __init__(self, root="data"):
        super(Elliptic, self).__init__(root, "Elliptic")


class Film(RD2CD):
    def __init__(self, root="data"):
        super(Film, self).__init__(root, "Film")


class Wiki(RD2CD):
    def __init__(self, root="data"):
        super(Wiki, self).__init__(root, "Wiki")


class Clothing(RD2CD):
    def __init__(self, root="data"):
        super(Clothing, self).__init__(root, "Clothing")


class Electronics(RD2CD):
    def __init__(self, root="data"):
        super(Electronics, self).__init__(root, "Electronics")


class Dblp(RD2CD):
    def __init__(self, root="data"):
        super(Dblp, self).__init__(root, "Dblp")


class Yelpchi(RD2CD):
    def __init__(self, root="data"):
        super(Yelpchi, self).__init__(root, "Yelpchi")


class Alpha(RD2CD):
    def __init__(self, root="data"):
        super(Alpha, self).__init__(root, "Alpha")


class Weibo(RD2CD):
    def __init__(self, root="data"):
        super(Weibo, self).__init__(root, "Weibo")


class bgp(RD2CD):
    def __init__(self, root="data"):
        super(bgp, self).__init__(root, "bgp")


class ssn5(RD2CD):
    def __init__(self, root="data"):
        super(ssn5, self).__init__(root, "ssn5")


class ssn7(RD2CD):
    def __init__(self, root="data"):
        super(ssn7, self).__init__(root, "ssn7")


class Aids(RD2CD):
    def __init__(self, root="data"):
        super(Aids, self).__init__(root, "Aids")


class Nba(RD2CD):
    def __init__(self, root="data"):
        super(Nba, self).__init__(root, "Nba")


class Pokec_z(RD2CD):
    def __init__(self, root="data"):
        super(Pokec_z, self).__init__(root, "Pokec_z")



import numpy as np
import torch
import torch.nn.functional as F
from scipy.linalg import expm
from cogdl.layers import GCNLayer

from .. import BaseModel


class GDC_GCN(BaseModel):
    r"""The GDC model from the `"Diffusion Improves Graph Learning"
    <https://arxiv.org/abs/1911.05485>`_ paper, with the PPR and heat matrix variants
    combined with GCN

    Args:
        num_features (int)  : Number of input features in ppr-preprocessed dataset.
        num_classes (int)   : Number of classes.
        hidden_size (int)   : The dimension of node representation.
        dropout (float)     : Dropout rate for model training.
        alpha (float)       : PPR polynomial filter param, 0 to 1.
        t (float)           : Heat polynomial filter param
        k (int)             : Top k nodes retained during sparsification.
        eps (float)         : Threshold for clipping.
        gdc_type (str)      : "none", "ppr", "heat"
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        # GCN Params
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--hidden-size", type=int, default=64)
        parser.add_argument("--dropout", type=float, default=0.5)

        # GDC Params
        parser.add_argument("--alpha", type=float, default=0.05)
        parser.add_argument("--t", type=float, default=5.0)
        parser.add_argument("--k", type=int, default=128)
        parser.add_argument("--eps", type=float, default=0.01)
        parser.add_argument("--dataset", default=None)
        parser.add_argument("--gdc-type", default="ppr")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.hidden_size,
            args.num_classes,
            args.dropout,
            args.alpha,
            args.t,
            args.k,
            args.eps,
            args.gdc_type,
        )

    def __init__(self, nfeat, nhid, nclass, dropout, alpha, t, k, eps, gdctype):
        super(GDC_GCN, self).__init__()

        # preproc params
        self.alpha = alpha
        self.t = t
        self.k = k
        self.eps = eps
        self.gdc_type = gdctype

        self.data = None

        # GCN init
        self.nfeat = nfeat
        self.gc1 = GCNLayer(nfeat, nhid)
        self.gc2 = GCNLayer(nhid, nclass)
        self.dropout = dropout

    def forward(self, graph):
        x = graph.x
        graph.sym_norm()

        x = F.dropout(x, self.dropout, training=self.training)
        x = F.relu(self.gc1(graph, x))

        x = F.dropout(x, self.dropout, training=self.training)
        x = self.gc2(graph, x)
        return x

    def predict(self, data=None):
        self.data.apply(lambda x: x.to(self.device))
        return self.forward(self.data)

    def reset_data(self, data):
        if self.data is None:
            data.to("cpu")
            self.data = self.preprocessing(data, gdc_type=self.gdc_type)
            data.to(self.device)

    def preprocessing(self, data, gdc_type="ppr"):
        # generate adjacency matrix from sparse representation

        def get_diffusion(x, edges):
            adj_matrix = self._get_adj_matrix(x, edges)

            if gdc_type == "none":
                print("No GDC filters chosen")
                processed_matrix = adj_matrix
            elif gdc_type == "ppr":
                print("PPR filters chosen")
                processed_matrix = self._get_ppr_matrix(adj_matrix, alpha=self.alpha)
            elif gdc_type == "heat":
                print("Heat filters chosen")
                processed_matrix = self._get_heat_matrix(adj_matrix, t=self.t)
            else:
                raise ValueError

            if gdc_type == "ppr" or gdc_type == "heat":
                if self.k:
                    print(f"Selecting top {self.k} edges per node.")
                    processed_matrix = self._get_top_k_matrix(processed_matrix, k=self.k)
                elif self.eps:
                    print(f"Selecting edges with weight greater than {self.eps}.")
                    processed_matrix = self._get_clipped_matrix(processed_matrix, eps=self.eps)
                else:
                    raise ValueError

            edges_i = []
            edges_j = []
            edge_attr = []
            for i, row in enumerate(processed_matrix):
                for j in np.where(row > 0)[0]:
                    edges_i.append(i)
                    edges_j.append(j)
                    edge_attr.append(processed_matrix[i, j])
            edge_index = [edges_i, edges_j]
            return torch.as_tensor(edge_index, dtype=torch.long), torch.as_tensor(edge_attr, dtype=torch.float)

        edge_index, edge_weight = get_diffusion(data.x, data.edge_index)
        data.edge_index = edge_index
        data.edge_weight = edge_weight

        if self.training and data._adj_train is not None:
            data.eval()
            edge_index, edge_weight = get_diffusion(data.x, data.edge_index)
            data.edge_index = edge_index
            data.edge_weight = edge_weight
            data.train()
        return data

    def _get_adj_matrix(self, x, edge_index):
        num_nodes = x.shape[0]
        adj_matrix = np.zeros(shape=(num_nodes, num_nodes))
        adj_matrix[edge_index[0], edge_index[1]] = 1.0
        return adj_matrix

    def _get_ppr_matrix(self, adj_matrix, alpha=0.1):
        num_nodes = adj_matrix.shape[0]
        A_tilde = adj_matrix + np.eye(num_nodes)
        D_tilde = np.diag(1 / np.sqrt(A_tilde.sum(axis=1)))
        H = D_tilde @ A_tilde @ D_tilde
        return alpha * np.linalg.inv(np.eye(num_nodes) - (1 - alpha) * H)

    def _get_heat_matrix(self, adj_matrix, t=5.0):
        num_nodes = adj_matrix.shape[0]
        A_tilde = adj_matrix + np.eye(num_nodes)
        D_tilde = np.diag(1 / np.sqrt(A_tilde.sum(axis=1)))
        H = D_tilde @ A_tilde @ D_tilde
        return expm(-t * (np.eye(num_nodes) - H))

    def _get_top_k_matrix(self, A, k=128):
        num_nodes = A.shape[0]
        row_idx = np.arange(num_nodes)
        A[A.argsort(axis=0)[: num_nodes - k], row_idx] = 0.0
        norm = A.sum(axis=0)
        norm[norm <= 0] = 1  # avoid dividing by zero
        return A / norm

    def _get_clipped_matrix(self, A, eps=0.01):
        A[A < eps] = 0.0
        norm = A.sum(axis=0)
        norm[norm <= 0] = 1  # avoid dividing by zero
        return A / norm



import torch
import torch.nn as nn
import torch.nn.functional as F

from cogdl.layers import DisenGCNLayer
from .. import BaseModel


class DisenGCN(BaseModel):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--hidden-size", type=int, default=64)
        parser.add_argument("--dropout", type=float, default=0.5)
        parser.add_argument("--K", type=int, nargs="+", default=[16, 8])
        parser.add_argument("--iterations", type=int, default=7)
        parser.add_argument("--tau", type=float, default=1)
        parser.add_argument("--activation", type=str, default="leaky_relu")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            in_feats=args.num_features,
            hidden_size=args.hidden_size,
            num_classes=args.num_classes,
            K=args.K,
            iterations=args.iterations,
            tau=args.tau,
            dropout=args.dropout,
            activation=args.activation,
        )

    def __init__(self, in_feats, hidden_size, num_classes, K, iterations, tau, dropout, activation):
        super(DisenGCN, self).__init__()
        self.K = K
        self.iterations = iterations
        self.dropout = dropout
        self.activation = activation
        self.num_layers = len(K)

        self.weight = nn.Parameter(torch.Tensor(hidden_size, num_classes))
        self.bias = nn.Parameter(torch.Tensor(num_classes))
        self.reset_parameters()

        shapes = [in_feats] + [hidden_size] * self.num_layers
        self.layers = nn.ModuleList(
            DisenGCNLayer(shapes[i], shapes[i + 1], K[i], iterations, tau, activation) for i in range(self.num_layers)
        )

    def reset_parameters(self):
        nn.init.xavier_normal_(self.weight.data, gain=1.414)
        nn.init.zeros_(self.bias.data)

    def forward(self, graph):
        h = graph.x
        graph.remove_self_loops()
        for layer in self.layers:
            h = layer(graph, h)
            # h = F.leaky_relu(h)
            h = F.dropout(h, p=self.dropout, training=self.training)
        out = torch.matmul(h, self.weight) + self.bias
        return out

    def predict(self, data):
        return self.forward(data)



import torch


CONFIGS = {
    "fast_spmm": None,
    "csrmhspmm": None,
    "csr_edge_softmax": None,
    "fused_gat_func": None,
    "fast_spmm_cpu": None,
    "spmm_flag": False,
    "mh_spmm_flag": False,
    "fused_gat_flag": False,
    "spmm_cpu_flag": False,
}


def check_fused_gat():
    return CONFIGS["fused_gat_func"] is not None


def initialize_spmm():
    if CONFIGS["spmm_flag"]:
        return
    CONFIGS["spmm_flag"] = True
    if torch.cuda.is_available():
        from cogdl.operators.spmm import csrspmm

        CONFIGS["fast_spmm"] = csrspmm
        # if csrspmm is None:
        #     print("Failed to load fast version of SpMM, use torch.scatter_add instead.")


def initialize_spmm_cpu():
    if CONFIGS["spmm_cpu_flag"]:
        return
    CONFIGS["spmm_cpu_flag"] = True

    from cogdl.operators.spmm import spmm_cpu

    CONFIGS["fast_spmm_cpu"] = spmm_cpu


def spmm_scatter(row, col, values, b):
    r"""
    Args:
        (row, col): Tensor, shape=(2, E)
        values : Tensor, shape=(E,)
        b : Tensor, shape=(N, d)
    """
    output = b.index_select(0, col) * values.unsqueeze(-1)
    output = torch.zeros_like(b).scatter_add_(0, row.unsqueeze(-1).expand_as(output), output)
    return output


def spmm_cpu(graph, x, fast_spmm_cpu=None):
    if fast_spmm_cpu is None:
        initialize_spmm_cpu()
        fast_spmm_cpu = CONFIGS["fast_spmm_cpu"]
    if fast_spmm_cpu is not None and str(x.device) == "cpu":
        if graph.out_norm is not None:
            x = graph.out_norm * x

        row_ptr, col_indices = graph.row_indptr, graph.col_indices
        csr_data = graph.raw_edge_weight
        x = fast_spmm_cpu(row_ptr.int(), col_indices.int(), csr_data, x)

        if graph.in_norm is not None:
            x = graph.in_norm * x
    else:
        row, col = graph.edge_index
        x = spmm_scatter(row, col, graph.edge_weight, x)
    return x


class SpMM_CPU(torch.nn.Module):
    def __init__(self):
        super().__init__()
        initialize_spmm_cpu()
        self.fast_spmm_cpu = CONFIGS["fast_spmm_cpu"]

    def forward(self, graph, x):
        return spmm_cpu(graph, x, self.fast_spmm_cpu)


def spmm(graph, x, actnn=False, fast_spmm=None):
    if fast_spmm is None:
        initialize_spmm()
        fast_spmm = CONFIGS["fast_spmm"]
    if fast_spmm is not None and str(x.device) != "cpu":
        if graph.out_norm is not None:
            x = graph.out_norm * x

        row_ptr, col_indices = graph.row_indptr, graph.col_indices
        csr_data = graph.raw_edge_weight
        x = fast_spmm(row_ptr.int(), col_indices.int(), x, csr_data, graph.is_symmetric(), actnn=actnn)

        if graph.in_norm is not None:
            x = graph.in_norm * x
    else:
        row, col = graph.edge_index
        x = spmm_scatter(row, col, graph.edge_weight, x)
    return x


class SpMM(torch.nn.Module):
    def __init__(self, actnn=False):
        super().__init__()
        initialize_spmm()
        self.actnn = actnn
        self.fast_spmm = CONFIGS["fast_spmm"]

    def forward(self, graph, x):
        return spmm(graph, x, self.actnn, self.fast_spmm)


def initialize_edge_softmax():
    if CONFIGS["mh_spmm_flag"]:
        return
    CONFIGS["mh_spmm_flag"] = True
    if torch.cuda.is_available():
        from cogdl.operators.edge_softmax import csr_edge_softmax
        from cogdl.operators.mhspmm import csrmhspmm

        CONFIGS["csrmhspmm"] = csrmhspmm
        CONFIGS["csr_edge_softmax"] = csr_edge_softmax


def edge_softmax_val(graph, edge_val):
    """
    Args:
        graph: cogdl.Graph
        edge_val: torch.Tensor, shape=(E, 1)
    Returns:
        Softmax values of edge values for nodes
    """
    edge_val_max = edge_val.max().item()
    while edge_val_max > 10:
        edge_val -= edge_val / 2
        edge_val_max = edge_val.max().item()

    with graph.local_graph():
        edge_val = torch.exp(edge_val)
        graph.edge_weight = edge_val
        x = torch.ones(graph.num_nodes, 1).to(edge_val.device)
        node_sum = spmm(graph, x).squeeze()
        row = graph.edge_index[0]
        softmax_values = edge_val / node_sum[row]
        return softmax_values


def edge_softmax(graph, edge_val, csr_edge_softmax=None):
    if csr_edge_softmax is None:
        initialize_edge_softmax()
        csr_edge_softmax = CONFIGS["csr_edge_softmax"]
    if csr_edge_softmax is not None and edge_val.device.type != "cpu":
        if len(edge_val.shape) == 1:
            edge_val = edge_val.view(-1, 1)
            val = csr_edge_softmax(graph.row_indptr.int(), edge_val)
            val = val.view(-1)
        else:
            val = csr_edge_softmax(graph.row_indptr.int(), edge_val)
        return val
    else:
        val = []
        for i in range(edge_val.shape[1]):
            val.append(edge_softmax_val(graph, edge_val[:, i]))
        return torch.stack(val).t()


class EdgeSoftmax(torch.nn.Module):
    def __init__(self):
        super().__init__()
        initialize_edge_softmax()
        self.csr_edge_softmax = CONFIGS["csr_edge_softmax"]

    def forward(self, graph, edge_val):
        return edge_softmax(graph, edge_val, self.csr_edge_softmax)


def mh_spmm(graph, attention, h, csrmhspmm=None, fast_spmm=None):
    if csrmhspmm is None:
        initialize_edge_softmax()
        csrmhspmm = CONFIGS["csrmhspmm"]
    nhead = h.shape[1]
    if csrmhspmm is not None and h.device.type != "cpu":
        if nhead > 1:
            h_prime = csrmhspmm(graph.row_indptr.int(), graph.col_indices.int(), h, attention)
            out = h_prime.view(h_prime.shape[0], -1)
        else:
            edge_weight = attention.view(-1)
            with graph.local_graph():
                graph.edge_weight = edge_weight
                out = spmm(graph, h.squeeze(1), fast_spmm=fast_spmm)
    else:
        with graph.local_graph():
            h_prime = []
            h = h.permute(1, 0, 2).contiguous()
            for i in range(nhead):
                edge_weight = attention[:, i]
                graph.edge_weight = edge_weight.contiguous()
                hidden = h[i]
                assert not torch.isnan(hidden).any()
                h_prime.append(spmm(graph, hidden, fast_spmm=fast_spmm))
        out = torch.cat(h_prime, dim=1)
    return out


class MultiHeadSpMM(torch.nn.Module):
    def __init__(self):
        super().__init__()
        initialize_spmm()
        initialize_edge_softmax()
        self.spmm = CONFIGS["fast_spmm"]
        self.csrmhspmm = CONFIGS["csrmhspmm"]

    def forward(self, graph, attention, h):
        return mh_spmm(graph, attention, h, csrmhspmm=self.csrmhspmm, fast_spmm=self.spmm)


def initialize_fused_gat():
    if CONFIGS["fused_gat_flag"]:
        return
    CONFIGS["fused_gat_flag"] = True
    if torch.cuda.is_available():
        from cogdl.operators.fused_gat import fused_gat_func

        CONFIGS["fused_gat_func"] = fused_gat_func


def fused_gat_op(attn_row, attn_col, graph, negative_slope, in_feat, fused_gat_func=None):
    if fused_gat_func is None:
        initialize_fused_gat()
        fused_gat_func = CONFIGS["fused_gat_func"]
    return fused_gat_func(
        attn_row,
        attn_col,
        graph.row_indptr.int(),
        graph.col_indices.int(),
        graph.row_indptr.int(),
        graph.col_indices.int(),
        negative_slope,
        in_feat,
    )


class FusedGATOp(torch.nn.Module):
    def __init__(self):
        super().__init__()
        initialize_fused_gat()
        self.fused_gat_func = CONFIGS["fused_gat_func"]

    def forward(self, attn_row, attn_col, graph, negative_slope, in_feat):
        return fused_gat_op(attn_row, attn_col, graph, negative_slope, in_feat, fused_gat_op=self.fused_gat_func)



import torch
import torch.nn as nn
import math
from cogdl.layers import AutoLayer
from cogdl.utils import get_activation

from .. import BaseModel


class AutoGNN(BaseModel):

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--num-classes", type=int)
        parser.add_argument("--num-layers", type=int, default=2)
        parser.add_argument("--hidden-size", type=int, default=64)
        parser.add_argument("--dropout", type=float, default=0.5)
        parser.add_argument("--residual", action="store_true")
        parser.add_argument("--norm", type=str, default=None)
        parser.add_argument("--activation", type=str, default="relu")
        parser.add_argument("--lambda", dest="lmbda", type=float, default=0.5)
        parser.add_argument("--alpha", type=float, default=0.1)
        parser.add_argument("--wd1", type=float, default=5e-4)
        parser.add_argument("--wd2", type=float, default=5e-4)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.hidden_size,
            args.num_classes,
            args.num_layers,
            args.dropout,
            args.activation,
            args.residual,
            args.norm,
            args.actnn,
            args.rp_ratio,
            args.alpha,
            args.lmbda,
            args.wd1,
            args.wd2,
        )

    def __init__(
        self,
        in_feats,
        hidden_size,
        out_feats,
        num_layers,
        dropout,
        activation="relu",
        residual=False,
        norm=None,
        actnn=False,
        rp_ratio=1,
        alpha=0.1,
        lmbda=0.5,
        wd1=5e-4,
        wd2=5e-4,
    ):
        super(AutoGNN, self).__init__()
        self.num_layers = num_layers
        self.alpha = alpha
        self.lmbda = lmbda
        self.wd1 = wd1
        self.wd2 = wd2

        self.layers = nn.ModuleList(
            [
                AutoLayer(
                    hidden_size,
                    hidden_size,
                    dropout=dropout,
                    residual=residual,
                    norm=norm,
                    activation=activation,
                    rp_ratio=rp_ratio,
                    alpha=alpha,
                    beta=math.log(self.lmbda / (i + 1) + 1),
                )
                for i in range(num_layers)
            ]
        )

        self.dropout = nn.Dropout(dropout)
        self.activation = get_activation(norm)

        self.fc_layers = nn.ModuleList()
        self.fc_layers.append(nn.Linear(in_feats, hidden_size))
        self.fc_layers.append(nn.Linear(hidden_size, out_feats))

        self.fc_parameters = list(self.fc_layers.parameters())
        self.conv_parameters = list(self.layers.parameters())

    def forward(self, graph):
        graph.sym_norm()
        x = graph.x
        h_init = self.dropout(x)
        h_init = self.activation(self.fc_layers[0](h_init))
        h = h_init

        for layer in self.layers:
            h = layer(graph, h, h_init)
        out = self.fc_layers[1](h)
        return out

    def get_optimizer(self, args):
        return torch.optim.Adam(
            [
                {"params": self.fc_parameters, "weight_decay": self.wd1},
                {"params": self.conv_parameters, "weight_decay": self.wd2},
            ],
            lr=args.lr,
        )



import os
import torch
from torch.utils.cpp_extension import load

path = os.path.join(os.path.dirname(__file__))

try:
    edge_softmax = load(
        name="edge_softmax",
        sources=[
            os.path.join(path, "edge_softmax/edge_softmax.cc"),
            os.path.join(path, "edge_softmax/edge_softmax.cu"),
        ],
        verbose=False,
    )

    def csr_edge_softmax(rowptr, h):
        return EdgeSoftmaxFunction.apply(rowptr, h)


except Exception:
    edge_softmax = None
    csr_edge_softmax = None


class EdgeSoftmaxFunction(torch.autograd.Function):
    @staticmethod
    def forward(ctx, rowptr, h):
        out = edge_softmax.edge_softmax(rowptr, h)
        ctx.save_for_backward(rowptr, out)
        return out

    @staticmethod
    def backward(ctx, grad_out):
        rowptr, out = ctx.saved_tensors
        grad_out = grad_out.contiguous()
        grad_softmax = edge_softmax.edge_softmax_backward(rowptr, out, grad_out)
        return None, grad_softmax



import torch

from .. import ModelWrapper


class GraphSAGEModelWrapper(ModelWrapper):
    def __init__(self, model, optimizer_cfg):
        super(GraphSAGEModelWrapper, self).__init__()
        self.model = model
        self.optimizer_cfg = optimizer_cfg

    def train_step(self, batch):
        x_src, y, adjs = batch
        pred = self.model(x_src, adjs)
        loss = self.default_loss_fn(pred, y)
        return loss

    def val_step(self, batch):
        x_src, y, adjs = batch
        pred = self.model(x_src, adjs)
        loss = self.default_loss_fn(pred, y)

        metric = self.evaluate(pred, y, metric="auto")

        self.note("val_loss", loss.item())
        self.note("val_metric", metric)

    def test_step(self, batch):
        dataset, test_loader = batch
        graph = dataset.data
        if hasattr(self.model, "inference"):
            pred = self.model.inference(graph.x, test_loader)
        else:
            pred = self.model(graph)
        pred = pred[graph.test_mask]
        y = graph.y[graph.test_mask]

        metric = self.evaluate(pred, y, metric="auto")
        self.note("test_loss", self.default_loss_fn(pred, y))
        self.note("test_metric", metric)

    def setup_optimizer(self):
        cfg = self.optimizer_cfg
        return torch.optim.Adam(self.parameters(), lr=cfg["lr"], weight_decay=cfg["weight_decay"])



import torch
import torch.nn as nn


class SELayer(nn.Module):
    """Squeeze-and-excitation networks"""

    def __init__(self, in_channels, se_channels):
        super(SELayer, self).__init__()

        self.in_channels = in_channels
        self.se_channels = se_channels

        self.encoder_decoder = nn.Sequential(
            nn.Linear(in_channels, se_channels),
            nn.ELU(),
            nn.Linear(se_channels, in_channels),
            nn.Sigmoid(),
        )

        # self.reset_parameters()

    def forward(self, x):
        """"""
        # Aggregate input representation
        x_global = torch.mean(x, dim=0)
        # Compute reweighting vector s
        s = self.encoder_decoder(x_global)

        return x * s



import math
import numpy as np
import scipy
import scipy.sparse as sp
from scipy.special import iv
from sklearn import preprocessing


class HeatKernel(object):
    def __init__(self, t=0.5, theta0=0.6, theta1=0.4):
        self.t = t
        self.theta0 = theta0
        self.theta1 = theta1

    def prop_adjacency(self, mx):
        mx_norm = preprocessing.normalize(mx.transpose(), "l1").transpose()
        adj = self.t * mx_norm
        adj.data = np.exp(adj.data)
        return adj / np.exp(self.t)

    def prop(self, mx, emb):
        adj = self.prop_adjacency(mx)
        return self.theta0 * emb + self.theta1 * adj.dot(emb)


class HeatKernelApproximation(object):
    def __init__(self, t=0.2, k=5):
        self.t = t
        self.k = k

    def taylor(self, mx, emb):
        mx_norm = preprocessing.normalize(mx, "l1")
        result = [math.exp(self.t) * emb]
        for i in range(self.k - 1):
            temp_mx = self.t * mx_norm.dot(result[-1]) / (i + 1)
            result.append(temp_mx)
        return sum(result)

    def chebyshev(self, mx, emb):
        mx = mx + sp.eye(emb.shape[0])
        mx = preprocessing.normalize(mx, "l1")
        conv = iv(0, self.t) * emb
        laplacian = sp.eye(emb.shape[0]) - mx
        Lx0 = emb
        Lx1 = laplacian.dot(emb)
        conv -= 2 * iv(1, self.t) * Lx1

        for i in range(2, self.k):
            Lx2 = 2 * laplacian.dot(Lx1) - Lx0
            conv += (-1) ** i * 2 * iv(i, self.t) * Lx2
            Lx0 = Lx1
            Lx1 = Lx2
        return conv

    def prop(self, mx, emb):
        return self.chebyshev(mx, emb)


class Gaussian(object):
    def __init__(self, mu=0.5, theta=1, rescale=False, k=3):
        self.theta = theta
        self.mu = mu
        self.k = k
        self.rescale = rescale
        self.coefs = [(-1) ** i * iv(i, self.theta) for i in range(k + 3)]
        self.coefs[0] = self.coefs[0] / 2

    # adj: 1 mul + 3 add,  emb: 2*k mul, 3*k add
    def prop(self, mx, emb):
        row_num, col_sum = mx.shape
        mx = mx + sp.eye(row_num)
        mx_norm = preprocessing.normalize(mx, "l1")
        mx_hat = (1 - self.mu) * sp.eye(row_num) - mx_norm

        Lx0 = emb
        Lx1 = mx_hat.dot(emb)
        Lx1 = 0.5 * mx_hat.dot(Lx1) - emb

        conv = iv(0, self.theta) * Lx0
        conv -= 2 * iv(1, self.theta) * Lx1
        for i in range(2, self.k):
            Lx2 = mx_hat.dot(Lx1)
            Lx2 = (mx_hat.dot(Lx2) - 2 * Lx1) - Lx0

            # Lx2 = 2 * mx_hat.dot(Lx1) - Lx0
            conv += (-1) ** i * 2 * iv(i, self.theta) * Lx2
            Lx0 = Lx1
            Lx1 = Lx2
        if self.rescale:
            conv = mx.dot(emb - conv)
        return conv


class PPR(object):
    """
    applying sparsification to accelerate computation
    """

    def __init__(self, alpha=0.5, k=10):
        self.alpha = alpha
        self.k = k
        self.alpha_list = [self.alpha * (1 - self.alpha) ** i for i in range(self.k)]
        self.epsilon = 1e-3

    def prop(self, mx, emb):
        mx_norm = preprocessing.normalize(mx, "l1")

        Lx = emb
        conv = self.alpha * Lx
        for i in range(1, self.k):
            Lx = (1 - self.alpha) * mx_norm.dot(Lx)
            conv += Lx
        return conv


class SignalRescaling(object):
    """
    - rescale signal of each node according to the degree of the node:
        - sigmoid(degree)
        - sigmoid(1/degree)
    """

    def __init__(self):
        pass

    def prop(self, mx, emb):
        mx = preprocessing.normalize(mx, "l1")
        degree = mx.sum(1).A.squeeze()

        degree_inv = 1.0 / degree
        signal_val = 1.0 / (1 + np.exp(-degree_inv))

        row_num, col_num = mx.shape
        q_ = sp.csc_matrix((signal_val, (np.arange(row_num), np.arange(col_num))), shape=(row_num, col_num))

        adj_norm = mx.dot(q_)
        adj_norm = preprocessing.normalize(adj_norm, "l1")
        conv = adj_norm.dot(emb)
        return conv


class ProNE(object):
    def __call__(self, A, a, order=10, mu=0.1, s=0.5):
        # NE Enhancement via Spectral Propagation
        print("Chebyshev Series -----------------")

        if order == 1:
            return a

        node_number = a.shape[0]

        A = sp.eye(node_number) + A
        DA = preprocessing.normalize(A, norm="l1")
        L = sp.eye(node_number) - DA

        M = L - mu * sp.eye(node_number)

        Lx0 = a
        Lx1 = M.dot(a)
        Lx1 = 0.5 * M.dot(Lx1) - a

        conv = iv(0, s) * Lx0
        conv -= 2 * iv(1, s) * Lx1
        for i in range(2, order):
            Lx2 = M.dot(Lx1)
            Lx2 = (M.dot(Lx2) - 2 * Lx1) - Lx0
            #         Lx2 = 2*L.dot(Lx1) - Lx0
            if i % 2 == 0:
                conv += 2 * iv(i, s) * Lx2
            else:
                conv -= 2 * iv(i, s) * Lx2
            Lx0 = Lx1
            Lx1 = Lx2
            del Lx2
        mm = A.dot(a - conv)
        return mm


class NodeAdaptiveEncoder(object):
    """
    - shrink negative values in signal/feature matrix
    - no learning
    """

    @staticmethod
    def prop(signal):
        mean_signal = signal.mean(1)
        mean_signal = 1.0 / (1 + np.exp(-mean_signal))
        sel_row, sel_col = np.where(signal < 0)
        mean_signal = mean_signal[sel_row]
        signal[sel_row, sel_col] = signal[sel_row, sel_col] * mean_signal
        return signal


def propagate(mx, emb, stype, space=None):
    if space is not None:
        if stype == "heat":
            heat_kernel = HeatKernelApproximation(t=space["t"])
            result = heat_kernel.prop(mx, emb)
        elif stype == "ppr":
            ppr = PPR(alpha=space["alpha"])
            result = ppr.prop(mx, emb)
        elif stype == "gaussian":
            gaussian = Gaussian(mu=space["mu"], theta=space["theta"])
            result = gaussian.prop(mx, emb)
        elif stype == "sc":
            signal_rs = SignalRescaling()
            result = signal_rs.prop(mx, emb)
        else:
            raise ValueError("please use filter in ['heat', 'ppr', 'gaussian', 'sc'], currently use {}".format(stype))
    else:
        if stype == "heat":
            heat_kernel = HeatKernelApproximation()
            result = heat_kernel.prop(mx, emb)
        elif stype == "ppr":
            ppr = PPR()
            result = ppr.prop(mx, emb)
        elif stype == "gaussian":
            gaussian = Gaussian()
            result = gaussian.prop(mx, emb)
        elif stype == "sc":
            signal_rs = SignalRescaling()
            result = signal_rs.prop(mx, emb)
        elif stype == "prone":
            signal_pro = ProNE()
            result = signal_pro(mx, emb)
        else:
            raise ValueError("please use filter in ['heat', 'ppr', 'gaussian', 'sc'], currently use {}".format(stype))
    return result


def get_embedding_dense(matrix, dimension):
    # get dense embedding via SVD
    U, s, Vh = scipy.linalg.svd(matrix, full_matrices=False, check_finite=False, overwrite_a=True)
    U = np.array(U)
    U = U[:, :dimension]
    s = s[:dimension]
    s = np.sqrt(s)
    U = U * s
    U = preprocessing.normalize(U, "l2")
    return U



import numpy as np
import torch
from torch.utils.data import DataLoader
from cogdl.data import Graph


class DataWrapper(object):
    @staticmethod
    def add_args(parser):
        pass

    def __init__(self, dataset=None):
        if dataset is not None:
            if hasattr(dataset, "get_loss_fn"):
                self.__loss_fn__ = dataset.get_loss_fn()
            if hasattr(dataset, "get_evaluator"):
                self.__evaluator__ = dataset.get_evaluator()
        else:
            self.__loss_fn__ = None
            self.__evaluator__ = None
        self.__dataset__ = dataset
        self.__training_data, self.__val_data, self.__test_data = None, None, None
        self.__num_training_data, self.__num_val_data, self.__num_test_data = 0, 0, 0
        self.__prepare_dataloader_per_epoch__ = False
        self.__back_to_cpu__ = False

    @property
    def data_back_to_cpu(self):
        return (
            self.__val_data is not None
            and self.__test_data is not None
            and (isinstance(self.__val_data.raw_data, Graph) or isinstance(self.__test_data.raw_data, Graph))
            and not isinstance(self.__training_data.raw_data, Graph)
        )

    def get_train_dataset(self):
        """
        Return the `wrapped` dataset for specific usage.
        For example, return `ClusteredDataset` in cluster_dw for DDP training.
        """
        raise NotImplementedError

    def get_val_dataset(self):
        """
        Similar to `self.get_train_dataset` but for validation.
        """
        raise NotImplementedError

    def get_test_dataset(self):
        """
        Similar to `self.get_train_dataset` but for test.
        """
        raise NotImplementedError

    def train_wrapper(self):
        """
        Return:
            1. DataLoader
            2. cogdl.Graph
            3. list of DataLoader or Graph
        Any other data formats other than DataLoader will not be traversed
        """
        pass

    def val_wrapper(self):
        pass

    def test_wrapper(self):
        pass

    def evaluation_wrapper(self):
        if self.__dataset__ is None:
            self.__dataset__ = getattr(self, "dataset", None)
        if self.__dataset__ is not None:
            return self.__dataset__

    def train_transform(self, batch):
        return batch

    def val_transform(self, batch):
        return batch

    def test_transform(self, batch):
        return batch

    def pre_transform(self):
        """Data Preprocessing before all runs"""
        pass

    def pre_stage(self, stage, model_w_out):
        """Processing before each run"""
        pass

    def post_stage(self, stage, model_w_out):
        """Processing after each run"""
        pass

    def refresh_per_epoch(self, name="train"):
        self.__prepare_dataloader_per_epoch__ = True

    def __refresh_per_epoch__(self):
        return self.__prepare_dataloader_per_epoch__

    def get_default_loss_fn(self):
        return self.__loss_fn__

    def get_default_evaluator(self):
        return self.__evaluator__

    def get_dataset(self):
        if self.__dataset__ is None:
            self.__dataset__ = getattr(self, "dataset", None)
        return self.__dataset__

    def prepare_training_data(self):
        train_data = self.train_wrapper()
        if train_data is not None:
            self.__training_data = OnLoadingWrapper(train_data, self.train_transform)

    def prepare_val_data(self):
        val_data = self.val_wrapper()
        if val_data is not None:
            self.__val_data = OnLoadingWrapper(val_data, self.val_transform)

    def prepare_test_data(self):
        test_data = self.test_wrapper()
        if test_data is not None:
            self.__test_data = OnLoadingWrapper(test_data, self.test_transform)

    def set_train_data(self, x):
        self.__training_data = x

    def set_val_data(self, x):
        self.__val_data = x

    def set_test_data(self, x):
        self.__test_data = x

    def on_train_wrapper(self):
        if self.__training_data is None:
            return None

        if self.__prepare_dataloader_per_epoch__:
            # TODO: reserve parameters for `prepare training data`
            self.prepare_training_data()
        return self.__training_data

    def on_val_wrapper(self):
        return self.__val_data

    def on_test_wrapper(self):
        return self.__test_data

    def train(self):
        if self.__dataset__ is None:
            self.__dataset__ = getattr(self, "dataset", None)
        if self.__dataset__ is not None and isinstance(self.__dataset__.data, Graph):
            self.__dataset__.data.train()

    def eval(self):
        if self.__dataset__ is None:
            self.__dataset__ = getattr(self, "dataset", None)
        if self.__dataset__ is not None and isinstance(self.__dataset__.data, Graph):
            self.__dataset__.data.eval()


class OnLoadingWrapper(object):
    def __init__(self, data, transform):
        """
        Args:
            data: `data` or `dataset`, that it, `cogdl.Graph` or `DataLoader`
        """
        self.raw_data = data
        self.data = self.__process_iterative_data__(data)
        self.__num_training_data = self.__get_min_len__(self.data)
        self.wrapped_data = self.__wrap_iteration__(self.data)
        self.ptr = 0
        self.transform = transform

    def __next__(self):
        if self.ptr < self.__num_training_data:
            self.ptr += 1
            batch = self.__next_batch__(self.wrapped_data)
            return self.transform(batch)
        else:
            self.ptr = 0
            # re-wrap the dataset per epoch
            self.wrapped_data = self.__wrap_iteration__(self.data)
            raise StopIteration

    def __iter__(self):
        return self

    def __len__(self):
        return self.__num_training_data

    def get_dataset_from_loader(self):
        return self.raw_data

    def __wrap_iteration__(self, inputs):
        # if isinstance(inputs, tuple):
        #     inputs = list(inputs)
        def iter_func(in_x):
            if isinstance(in_x, list) or isinstance(in_x, DataLoader):
                for item in in_x:
                    yield item
            else:
                yield in_x

        if isinstance(inputs, list):
            outputs = [None] * len(inputs)
            for i, item in enumerate(inputs):
                outputs[i] = self.__wrap_iteration__(item)
        elif isinstance(inputs, dict):
            outputs = {key: None for key in inputs.keys()}
            for key, val in inputs.items():
                outputs[key] = self.__wrap_iteration__(val)
        else:
            # return LoaderWrapper(inputs)
            return iter_func(inputs)
        return outputs

    def __process_iterative_data__(self, inputs):
        if inputs is None:
            return None
        # if isinstance(inputs, tuple):
        #     inputs = list(inputs)

        if isinstance(inputs, list):
            for i, item in enumerate(inputs):
                inputs[i] = self.__process_iterative_data__(item)
        elif isinstance(inputs, dict):
            for key, val in inputs.items():
                inputs[key] = self.__process_iterative_data__(val)
        else:
            # return self.__batch_wrapper__(inputs)
            return inputs
        return inputs

    def __next_batch__(self, inputs):
        # if isinstance(inputs, tuple):
        #     inputs = list(inputs)

        if isinstance(inputs, list):
            outputs = [None] * len(inputs)
            for i, item in enumerate(inputs):
                outputs[i] = self.__next_batch__(item)
        elif isinstance(inputs, dict):
            outputs = {key: None for key in inputs.keys()}
            for key, val in inputs.items():
                outputs[key] = self.__next_batch__(val)
        else:
            return next(inputs)
        return outputs

    def __get_min_len__(self, inputs):
        if inputs is None:
            return None

        # if isinstance(inputs, tuple):
        #     inputs = list(inputs)
        if isinstance(inputs, list):
            outputs = [0] * len(inputs)
            for i, item in enumerate(inputs):
                inputs[i] = self.__get_min_len__(item)
            return np.min(outputs)
        # elif isinstance(inputs, dict):
        #     outputs = {key: 0 for key in inputs.keys()}
        #     for i, val in enumerate(inputs.values()):
        #         outputs[i] = self.__get_min_len__(val)
        #     return np.min(list(outputs.values()))
        else:
            if isinstance(inputs, DataLoader):
                return len(inputs)
            else:
                return 1



import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from cogdl.layers import GCNIILayer
from .. import BaseModel


class GCNII(BaseModel):
    """
    Implementation of GCNII in paper `"Simple and Deep Graph Convolutional Networks" <https://arxiv.org/abs/2007.02133>`_.

    Parameters
    -----------
    in_feats : int
        Size of each input sample
    hidden_size : int
        Size of each hidden unit
    out_feats : int
        Size of each out sample
    num_layers : int
    dropout : float
    alpha : float
        Parameter of initial residual connection
    lmbda : float
        Parameter of identity mapping
    wd1 : float
        Weight-decay for Fully-connected layers
    wd2 : float
        Weight-decay for convolutional layers
    """

    @staticmethod
    def add_args(parser):
        parser.add_argument("--hidden-size", type=int, default=64)
        parser.add_argument("--num-layers", type=int, default=64)
        parser.add_argument("--lambda", dest="lmbda", type=float, default=0.5)
        parser.add_argument("--alpha", type=float, default=0.1)
        parser.add_argument("--dropout", type=float, default=0.6)
        parser.add_argument("--wd1", type=float, default=5e-4)
        parser.add_argument("--wd2", type=float, default=5e-4)
        parser.add_argument("--residual", action="store_true")

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            in_feats=args.num_features,
            hidden_size=args.hidden_size,
            out_feats=args.num_classes,
            num_layers=args.num_layers,
            dropout=args.dropout,
            alpha=args.alpha,
            lmbda=args.lmbda,
            wd1=args.wd1,
            wd2=args.wd2,
            residual=args.residual,
            actnn=args.actnn,
        )

    def __init__(
        self,
        in_feats,
        hidden_size,
        out_feats,
        num_layers,
        dropout=0.5,
        alpha=0.1,
        lmbda=1,
        wd1=0.0,
        wd2=0.0,
        residual=False,
        actnn=False,
    ):
        super(GCNII, self).__init__()
        Layer = GCNIILayer
        Linear = nn.Linear
        Dropout = nn.Dropout
        ReLU = nn.ReLU(inplace=True)
        if actnn:
            try:
                from cogdl.layers.actgcnii_layer import ActGCNIILayer
                from actnn.layers import QLinear, QReLU, QDropout
            except Exception:
                print("Please install the actnn library first.")
                exit(1)
            Layer = ActGCNIILayer
            Linear = QLinear
            Dropout = QDropout
            ReLU = QReLU()

        self.fc_layers = nn.ModuleList()
        self.fc_layers.append(Linear(in_feats, hidden_size))
        self.fc_layers.append(Linear(hidden_size, out_feats))

        self.dropout = Dropout(dropout)
        self.activation = ReLU
        self.alpha = alpha
        self.lmbda = lmbda
        self.wd1 = wd1
        self.wd2 = wd2

        self.layers = nn.ModuleList(
            Layer(hidden_size, self.alpha, math.log(self.lmbda / (i + 1) + 1), residual) for i in range(num_layers)
        )

        self.fc_parameters = list(self.fc_layers.parameters())
        self.conv_parameters = list(self.layers.parameters())

    def forward(self, graph):
        graph.sym_norm()
        x = graph.x
        init_h = self.dropout(x)
        init_h = self.activation(self.fc_layers[0](init_h))

        h = init_h

        for layer in self.layers:
            h = self.dropout(h)
            h = layer(graph, h, init_h)
            h = self.activation(h)
        h = self.dropout(h)
        out = self.fc_layers[1](h)
        return out

    def predict(self, graph):
        return self.forward(graph)

    def get_optimizer(self, args):
        return torch.optim.Adam(
            [
                {"params": self.fc_parameters, "weight_decay": self.wd1},
                {"params": self.conv_parameters, "weight_decay": self.wd2},
            ],
            lr=args.lr,
        )



import re

import torch
from cogdl.data import Graph, Adjacency


def batch_graphs(graphs):
    return Batch.from_data_list(graphs, class_type=Graph)


class Batch(Graph):
    r"""A plain old python object modeling a batch of graphs as one big
    (dicconnected) graph. With :class:`cogdl.data.Data` being the
    base class, all its methods can also be used here.
    In addition, single graphs can be reconstructed via the assignment vector
    :obj:`batch`, which maps each node to its respective graph identifier.
    """

    def __init__(self, batch=None, **kwargs):
        super(Batch, self).__init__(**kwargs)
        self.batch = batch
        self.__data_class__ = Graph
        self.__slices__ = None

    @staticmethod
    def from_data_list(data_list, class_type=None):
        r"""Constructs a batch object from a python list holding
        :class:`cogdl.data.Data` objects.
        The assignment vector :obj:`batch` is created on the fly.
        Additionally, creates assignment batch vectors for each key in
        :obj:`follow_batch`."""

        # keys = [set(data.keys) for data in data_list]
        keys = [set(data.keys) for data in data_list]
        keys = list(set.union(*keys))
        assert "batch" not in keys

        if class_type is not None:
            batch = class_type()
        else:
            batch = Batch()
            batch.__data_class__ = data_list[0].__class__
        batch.__slices__ = {key: [0] for key in keys}

        for key in keys:
            batch[key] = []

        # for key in follow_batch:
        #     batch["{}_batch".format(key)] = []

        cumsum = {key: 0 for key in keys}
        batch.batch = []
        num_nodes_cum = [0]
        num_edges_cum = [0]
        for i, data in enumerate(data_list):
            for key in data.keys:
                item = data[key]
                if torch.is_tensor(item) and item.dtype != torch.bool:
                    item = item + cumsum[key]
                if torch.is_tensor(item):
                    size = item.size(data.cat_dim(key, data[key]))
                else:
                    size = 1
                batch.__slices__[key].append(size + batch.__slices__[key][-1])
                cumsum[key] = cumsum[key] + data.__inc__(key, item)
                batch[key].append(item)

                # if key in follow_batch:
                #     item = torch.full((size,), i, dtype=torch.long)
                #     batch["{}_batch".format(key)].append(item)

            num_nodes = data.num_nodes
            if num_nodes is not None:
                num_nodes_cum.append(num_nodes + num_nodes_cum[-1])
                num_edges_cum.append(data.num_edges + num_edges_cum[-1])
                item = torch.full((num_nodes,), i, dtype=torch.long)
                batch.batch.append(item)
        if num_nodes is None:
            batch.batch = None
        for key in batch.keys:
            item = batch[key][0]
            if torch.is_tensor(item):
                batch[key] = torch.cat(batch[key], dim=data_list[0].cat_dim(key, item))
            elif isinstance(item, int) or isinstance(item, float):
                batch[key] = torch.tensor(batch[key])
            elif isinstance(item, Adjacency):
                target = Adjacency()
                for k in item.keys:
                    if item[k] is None:
                        continue
                    if k == "row" or k == "col":
                        _item = torch.cat(
                            [x[k] + num_nodes_cum[i] for i, x in enumerate(batch[key])], dim=item.cat_dim(k, None)
                        )
                    elif k == "row_ptr":
                        _item = torch.cat(
                            [x[k][:-1] + num_edges_cum[i] for i, x in enumerate(batch[key][:-1])],
                            dim=item.cat_dim(k, None),
                        )
                        _item = torch.cat([_item, batch[key][-1][k] + num_edges_cum[-2]], dim=item.cat_dim(k, None))
                    else:
                        _item = torch.cat([x[k] for i, x in enumerate(batch[key])], dim=item.cat_dim(k, None))
                    target[k] = _item
                batch[key] = target.to(item.device)
        return batch.contiguous()

    def cumsum(self, key, item):
        r"""If :obj:`True`, the attribute :obj:`key` with content :obj:`item`
        should be added up cumulatively before concatenated together.

        .. note::

            This method is for internal use only, and should only be overridden
            if the batch concatenation process is corrupted for a specific data
            attribute.
        """
        return bool(re.search("(index|face)", key))

    @property
    def num_graphs(self):
        """Returns the number of graphs in the batch."""
        return self.batch[-1].item() + 1



class Logger:
    def __init__(self, log_path):
        self.log_path = log_path

    def start(self):
        pass

    def note(self, metrics, step=None):
        pass

    def finish(self):
        pass



import collections
import os.path as osp
from itertools import repeat
import numpy as np

import torch.utils.data

from cogdl.data import Adjacency, Graph
from cogdl.utils import makedirs
from cogdl.utils import Accuracy, CrossEntropyLoss


def to_list(x):
    if not isinstance(x, collections.Iterable) or isinstance(x, str):
        x = [x]
    return x


def files_exist(files):
    return all([osp.exists(f) for f in files])


class Dataset(torch.utils.data.Dataset):
    r"""Dataset base class for creating graph datasets.
    See `here <https://rusty1s.github.io/pycogdl/build/html/notes/
    create_dataset.html>`__ for the accompanying tutorial.

    Args:
        root (string): Root directory where the dataset should be saved.
        transform (callable, optional): A function/transform that takes in an
            :obj:`cogdl.data.Data` object and returns a transformed
            version. The data object will be transformed before every access.
            (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in
            an :obj:`cogdl.data.Data` object and returns a
            transformed version. The data object will be transformed before
            being saved to disk. (default: :obj:`None`)
        pre_filter (callable, optional): A function that takes in an
            :obj:`cogdl.data.Data` object and returns a boolean
            value, indicating whether the data object should be included in the
            final dataset. (default: :obj:`None`)
    """

    @staticmethod
    def add_args(parser):
        """Add dataset-specific arguments to the parser."""
        pass

    @property
    def raw_file_names(self):
        r"""The name of the files to find in the :obj:`self.raw_dir` folder in
        order to skip the download."""
        raise NotImplementedError

    @property
    def processed_file_names(self):
        r"""The name of the files to find in the :obj:`self.processed_dir`
        folder in order to skip the processing."""
        raise NotImplementedError

    def download(self):
        r"""Downloads the dataset to the :obj:`self.raw_dir` folder."""
        raise NotImplementedError

    def process(self):
        r"""Processes the dataset to the :obj:`self.processed_dir` folder."""
        raise NotImplementedError

    def __len__(self):
        r"""The number of examples in the dataset."""
        return 1

    def get(self, idx):
        r"""Gets the data object at index :obj:`idx`."""
        raise NotImplementedError

    def __init__(self, root, transform=None, pre_transform=None, pre_filter=None):
        super(Dataset, self).__init__()

        self.root = osp.expanduser(osp.normpath(root))
        self.raw_dir = osp.join(self.root, "raw")
        self.processed_dir = osp.join(self.root, "processed")
        self.transform = transform
        self.pre_transform = pre_transform
        self.pre_filter = pre_filter

        self._download()
        self._process()

    @property
    def num_features(self):
        r"""Returns the number of features per node in the graph."""
        if hasattr(self, "data") and isinstance(self.data, Graph):
            return self.data.num_features
        elif hasattr(self, "data") and isinstance(self.data, list):
            return self.data[0].num_features
        else:
            return 0

    @property
    def raw_paths(self):
        r"""The filepaths to find in order to skip the download."""
        files = to_list(self.raw_file_names)
        return [osp.join(self.raw_dir, f) for f in files]

    @property
    def processed_paths(self):
        r"""The filepaths to find in the :obj:`self.processed_dir`
        folder in order to skip the processing."""
        files = to_list(self.processed_file_names)
        return [osp.join(self.processed_dir, f) for f in files]

    def _download(self):
        if files_exist(self.raw_paths):  # pragma: no cover
            return

        makedirs(self.raw_dir)
        self.download()

    def _process(self):
        if files_exist(self.processed_paths):  # pragma: no cover
            return

        print("Processing...")

        makedirs(self.processed_dir)
        self.process()

        print("Done!")

    def get_evaluator(self):
        return Accuracy()

    def get_loss_fn(self):
        return CrossEntropyLoss()

    def __getitem__(self, idx):  # pragma: no cover
        r"""Gets the data object at index :obj:`idx` and transforms it (in case
        a :obj:`self.transform` is given)."""
        assert idx == 0
        data = self.data
        data = data if self.transform is None else self.transform(data)
        return data

    @property
    def num_classes(self):
        r"""The number of classes in the dataset."""
        if hasattr(self, "y") and self.y is not None:
            y = self.y
        elif hasattr(self, "data") and hasattr(self.data, "y") and self.data.y is not None:
            y = self.data.y
        else:
            return 0
        return y.max().item() + 1 if y.dim() == 1 else y.size(1)

    @property
    def edge_attr_size(self):
        return None

    @property
    def max_degree(self):
        return self.data.degrees().max().item() + 1

    @property
    def max_graph_size(self):
        return self.data.num_nodes

    @property
    def num_graphs(self):
        return 1

    def __repr__(self):  # pragma: no cover
        return "{}".format(self.__class__.__name__)


class MultiGraphDataset(Dataset):
    def __init__(self, root=None, transform=None, pre_transform=None, pre_filter=None):
        super(MultiGraphDataset, self).__init__(root, transform, pre_transform, pre_filter)
        self.data, self.slices = None, None

    @property
    def num_classes(self):
        if hasattr(self, "y"):
            y = self.y
        elif hasattr(self, "data") and hasattr(self.data[0], "y"):
            y = torch.cat([x.y for x in self.data], dim=0)
        else:
            return 0
        return y.max().item() + 1 if y.dim() == 1 else y.size(1)

    @property
    def num_features(self):
        if isinstance(self[0], Graph):
            return self[0].num_features
        else:
            return 0

    @property
    def max_degree(self):
        max_degree = [x.degrees().max().item() for x in self.data]
        max_degree = np.max(max_degree) + 1
        return max_degree

    @property
    def num_graphs(self):
        return len(self.data)

    @property
    def max_graph_size(self):
        return np.max([g.num_nodes for g in self.data])

    def len(self):
        if isinstance(self.data, list):
            return len(self.data)
        else:
            for item in self.slices.values():
                return len(item) - 1
            return 0

    def _get(self, idx):
        data = self.data.__class__()

        if hasattr(self.data, "__num_nodes__"):
            data.num_nodes = self.data.__num_nodes__[idx]

        for key in self.data.__old_keys__():
            item, slices = self.data[key], self.slices[key]
            # start, end = slices[idx].item(), slices[idx + 1].item()
            start, end = int(slices[idx]), int(slices[idx + 1])
            if key == "edge_index":
                data[key] = (item[0][start:end], item[1][start:end])
            else:
                if torch.is_tensor(item):
                    s = list(repeat(slice(None), item.dim()))
                    s[self.data.__cat_dim__(key, item)] = slice(start, end)
                elif start + 1 == end:
                    s = slices[start]
                else:
                    s = slice(start, end)
                data[key] = item[s]
        return data

    def get(self, idx):
        try:
            idx = int(idx)
        except Exception:
            idx = idx
        if torch.is_tensor(idx):
            idx = idx.numpy().tolist()
        if isinstance(idx, int):
            if self.slices is not None:
                return self._get(idx)
            return self.data[idx]
        if isinstance(idx, slice):
            start = idx.start
            end = idx.stop
            step = idx.step if idx.step else 1
            idx = list(range(start, end, step))

        if len(idx) > 1:
            # unsupport `slice`
            if self.slices is not None:
                return [self._get(int(i)) for i in idx]
            return [self.data[i] for i in idx]

    def __getitem__(self, item):
        return self.get(item)

    def __len__(self):
        return len(self.data)

    def __repr__(self):  # pragma: no cover
        return "{}({})".format(self.__class__.__name__, len(self))
